#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# test_ongoing.py
# Run unit tests for ongoing entries in Pygenda
#
# Copyright (C) 2023 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.
#

import unittest
from datetime import date, datetime, timedelta
from dateutil import tz
from os import remove as os_remove

# Add '..' to path, so this can be run from test directory
import sys
sys.path.append('..')

# Import the modules we need for testing...
from pygenda.pygenda_calendar import Calendar
from pygenda.pygenda_config import Config
from pygenda.pygenda_entryinfo import EntryInfo
from pygenda.pygenda_util import get_local_tz, _set_local_tz as set_local_tz


class TestOngoing(unittest.TestCase):
    maxDiff = None # show unlimited chars when showing diffs
    TESTFILE_NAME = 'test_ongoing_TESTFILE.ics'

    @classmethod
    def setUpClass(cls):
        # Called once before all tests
        # Save local timezone, so running tests will test in your tz
        cls.saved_tz = get_local_tz()
        # Override config options so it uses our test ics file
        Config.set('calendar', 'type', 'icalfile')
        Config.set('calendar', 'filename', cls.TESTFILE_NAME)
        Config.set('calendar', 'display_name', 'Test calendar for test_ongoing')
        Config.set('calendar', 'readonly', None)
        Config.set('calendar', 'entry_type', None)
        Config.set('calendar1', 'type', None) # so only specified file opened


    def setUp(self) -> None:
        # This is called before each individual test function

        # Delete test ics files, so a new one is created
        self._delete_testfile()

        # Reset timezone for next test (might be changed in test)
        set_local_tz(self.saved_tz)

        # Call init to create the calendar
        Calendar.init()


    @classmethod
    def tearDownClass(cls) -> None:
        # This is called after final test
        cls._delete_testfile()


    @classmethod
    def _delete_testfile(cls) -> None:
        # Helper function for setup/teardown
        try:
            os_remove(cls.TESTFILE_NAME)
        except FileNotFoundError:
            pass


    #@unittest.skip
    def test_01_empty(self) -> None:
        # Test empty database
        self.check_ongoing_count(date(1963,11,27), 0)
        self.check_ongoing_count(date(1999,12,31), 0)
        self.check_ongoing_count(date(2000,1,1), 0)
        self.check_ongoing_count(date(2020,1,1), 0)
        self.check_ongoing_count(date(2021,1,1), 0)
        self.check_ongoing_count(date(2020,2,28), 0)
        self.check_ongoing_count(date(2020,2,29), 0)
        self.check_ongoing_count(date(2020,3,1), 0)


    #@unittest.skip
    def test_02_untimed(self) -> None:
        # Test untimed, single-day and multi-day events

        # Simple untimed event
        Calendar.new_entry(EntryInfo(desc='untimed', start_dt=date(2020,2,3)))
        self.check_ongoing_count(date(2020,2,3), 0)
        self.check_ongoing_count(date(2020,2,4), 0)

        # one-day event
        Calendar.new_entry(EntryInfo(desc='one-day', start_dt=date(2020,2,10), end_dt=date(2020,2,11)))
        self.check_ongoing_count(date(2020,2,10), 0)
        self.check_ongoing_count(date(2020,2,11), 0)

        # two-day event
        Calendar.new_entry(EntryInfo(desc='two-day', start_dt=date(2020,2,16), end_dt=date(2020,2,18)))
        self.check_ongoing_count(date(2020,2,16), 0)
        self.check_ongoing_count(date(2020,2,17), 1)
        self.check_ongoing_count(date(2020,2,18), 0)


    #@unittest.skip
    def test_03_timed(self) -> None:
        # Test timed events, with duration & with endtime

        # Simple timed event
        Calendar.new_entry(EntryInfo(desc='timed', start_dt=datetime(2021,3,4,10,30)))
        self.check_ongoing_count(date(2021,3,4), 0)
        self.check_ongoing_count(date(2021,3,5), 0)

        # Timed event, short duration (doesn't cross midnight)
        Calendar.new_entry(EntryInfo(desc='timed short duration', start_dt=datetime(2021,3,8,11,45), duration=timedelta(minutes=30)))
        self.check_ongoing_count(date(2021,3,8), 0)
        self.check_ongoing_count(date(2021,3,9), 0)

        # Timed event, medium duration (once across midnight)
        Calendar.new_entry(EntryInfo(desc='timed med duration', start_dt=datetime(2021,3,13,19,30), duration=timedelta(hours=10)))
        self.check_ongoing_count(date(2021,3,13), 0)
        self.check_ongoing_count(date(2021,3,14), 1)
        self.check_ongoing_count(date(2021,3,15), 0)

        # Timed event, long duration (cross midnight multiple times)
        Calendar.new_entry(EntryInfo(desc='timed long duration', start_dt=datetime(2021,3,19,13,0), duration=timedelta(days=15, hours=13)))
        self.check_ongoing_count(date(2021,3,19), 0)
        dt = date(2021,3,20)
        for i in range(16):
            self.check_ongoing_count(dt, 1)
            dt += timedelta(days=1)
        self.check_ongoing_count(date(2021,4,4), 1) # also checked in loop
        self.check_ongoing_count(date(2021,4,5), 0)

        # Timed event, endtime short (doesn't cross midnight)
        Calendar.new_entry(EntryInfo(desc='timed short endtime', start_dt=datetime(2021,4,8,10,40), end_dt=datetime(2021,4,8,18,49)))
        self.check_ongoing_count(date(2021,4,8), 0)
        self.check_ongoing_count(date(2021,4,9), 0)

        # Timed event, endtime medium (once across midnight)
        Calendar.new_entry(EntryInfo(desc='timed med endtime', start_dt=datetime(2021,4,13,18,15), end_dt=datetime(2021,4,14,18,15)))
        self.check_ongoing_count(date(2021,4,13), 0)
        self.check_ongoing_count(date(2021,4,14), 1)
        self.check_ongoing_count(date(2021,4,15), 0)

        # Timed event, long duration (cross midnight multiple times)
        Calendar.new_entry(EntryInfo(desc='timed long endtime', start_dt=datetime(2021,4,19,13,0), end_dt=datetime(2021,5,1,2,0)))
        self.check_ongoing_count(date(2021,4,19), 0)
        dt = date(2021,4,20)
        for i in range(12):
            self.check_ongoing_count(dt, 1)
            dt += timedelta(days=1)
        self.check_ongoing_count(date(2021,5,1), 1) # also checked in loop
        self.check_ongoing_count(date(2021,5,2), 0)


    #@unittest.skip
    def test_04_overlapping(self) -> None:
        # Test overlapping events
        Calendar.new_entry(EntryInfo(desc='overlapping 1', start_dt=datetime(2021,5,25), end_dt=datetime(2021,6,4)))
        Calendar.new_entry(EntryInfo(desc='overlapping 2', start_dt=datetime(2021,5,27,23,59), duration=timedelta(days=10)))
        Calendar.new_entry(EntryInfo(desc='overlapping 3', start_dt=date(2021,5,28)))
        Calendar.new_entry(EntryInfo(desc='overlapping 4', start_dt=datetime(2021,6,1,17,45)))
        Calendar.new_entry(EntryInfo(desc='overlapping 5', start_dt=datetime(2021,5,30,19,0), end_dt=datetime(2021,5,31,0,0)))
        self.check_ongoing_count(date(2021,5,25), 0)
        dt = date(2021,5,26)
        for i in range(2): # 26,27
            self.check_ongoing_count(dt, 1)
            dt += timedelta(days=1)
        for i in range(7): # 28,29,30,31,1,2,3
            self.check_ongoing_count(dt, 2)
            dt += timedelta(days=1)
        for i in range(3): # 4,5,6
            self.check_ongoing_count(dt, 1)
            dt += timedelta(days=1)
        self.check_ongoing_count(date(2021,6,6), 1) # already checked in loop
        self.check_ongoing_count(date(2021,6,7), 0)


    #@unittest.skip
    def test_05_midnight(self) -> None:
        # Test cases starting/ending at midnight
        # Timed event at midnight
        Calendar.new_entry(EntryInfo(desc='at midnight', start_dt=datetime(2021,6,9,0,0)))
        self.check_ongoing_count(date(2021,6,9), 0)
        self.check_ongoing_count(date(2021,6,10), 0)

        # Event ending at midnight
        Calendar.new_entry(EntryInfo(desc='ends midnight', start_dt=datetime(2021,6,12,19,10), end_dt=datetime(2021,6,13,0,0)))
        self.check_ongoing_count(date(2021,6,12), 0)
        self.check_ongoing_count(date(2021,6,13), 0)

        # Multi-day event ending at midnight
        Calendar.new_entry(EntryInfo(desc='ends midnight in two days', start_dt=datetime(2021,6,16,14,0,0), end_dt=datetime(2021,6,19,0,0)))
        self.check_ongoing_count(date(2021,6,16), 0)
        self.check_ongoing_count(date(2021,6,17), 1)
        self.check_ongoing_count(date(2021,6,18), 1)
        self.check_ongoing_count(date(2021,6,19), 0)

        # Event with duration ending at midnight
        Calendar.new_entry(EntryInfo(desc='duration to midnight', start_dt=datetime(2021,6,21,17,30), duration=timedelta(hours=6, minutes=30)))
        self.check_ongoing_count(date(2021,6,21), 0)
        self.check_ongoing_count(date(2021,6,22), 0)

        # Multi-day event with duration ending at midnight
        Calendar.new_entry(EntryInfo(desc='duration to midnight', start_dt=datetime(2021,6,25,11,15), duration=timedelta(days=2, hours=12, minutes=45)))
        self.check_ongoing_count(date(2021,6,25), 0)
        self.check_ongoing_count(date(2021,6,26), 1)
        self.check_ongoing_count(date(2021,6,27), 1)
        self.check_ongoing_count(date(2021,6,28), 0)


    #@unittest.skip
    def test_06_timezone(self) -> None:
        # Test cases with non-local timezone

        # Just has a timezone, to check it even works
        tz_LON = tz.gettz('Europe/London')
        self.assertTrue(tz_LON) # check not None
        Calendar.new_entry(EntryInfo(desc='with timezone', start_dt=datetime(2021,7,4,12,0,tzinfo=tz_LON)))
        self.check_ongoing_count(date(2021,7,4), 0)
        self.check_ongoing_count(date(2021,7,5), 0)

        # With an endtime, not crossing midnight
        Calendar.new_entry(EntryInfo(desc='with timezone & endtime', start_dt=datetime(2021,7,6,12,0,tzinfo=tz_LON), end_dt=datetime(2021,7,6,13,0,tzinfo=tz_LON)))
        self.check_ongoing_count(date(2021,7,6), 0)
        self.check_ongoing_count(date(2021,7,7), 0)

        # With an endtime, crossing midnight
        Calendar.new_entry(EntryInfo(desc='with timezone, crossing midnight', start_dt=datetime(2021,7,8,12,0,tzinfo=tz_LON), end_dt=datetime(2021,7,9,11,59,tzinfo=tz_LON)))
        self.check_ongoing_count(date(2021,7,8), 0)
        self.check_ongoing_count(date(2021,7,9), 1)
        self.check_ongoing_count(date(2021,7,10), 0)

        # With an endtime, crossing midnight locally
        tz_SAM = tz.gettz('Pacific/Samoa') # UTC-11
        self.assertTrue(tz_SAM) # check not None
        Calendar.new_entry(EntryInfo(desc='with westerly non-local timezone & endtime', start_dt=datetime(2021,7,10,0,5,tzinfo=tz_SAM), end_dt=datetime(2021,7,10,23,55,tzinfo=tz_SAM)))
        self.check_ongoing_count(date(2021,7,10), 0)
        self.check_ongoing_count(date(2021,7,11), 1)
        self.check_ongoing_count(date(2021,7,12), 0)

        # We want to test somewhere that is ahead of local time, so
        # let's take the most easterly timezone we know of.
        tz_NZ = tz.gettz('Pacific/Auckland')
        self.assertTrue(tz_NZ) # check not None

        # Check to see if local timezone is same as event timezone,
        # and if so change local tz to something "neutral".
        if get_local_tz().utcoffset(datetime(2021,7,12,12))>=timedelta(hours=12):
            set_local_tz(tz_LON)

        Calendar.new_entry(EntryInfo(desc='with easterly non-local timezone & endtime', start_dt=datetime(2021,7,13,0,5,tzinfo=tz_NZ), end_dt=datetime(2021,7,13,23,55,tzinfo=tz_NZ)))
        self.check_ongoing_count(date(2021,7,12), 0)
        self.check_ongoing_count(date(2021,7,13), 1)
        self.check_ongoing_count(date(2021,7,14), 0)


    #@unittest.skip
    def test_07_timezone_duration(self) -> None:
        # Test cases with non-local timezone using duration

        tz_LON = tz.gettz('Europe/London')
        self.assertTrue(tz_LON) # check not None

        # With a duration, not crossing midnight
        Calendar.new_entry(EntryInfo(desc='with timezone & duration', start_dt=datetime(2021,7,6,12,0,tzinfo=tz_LON), duration=timedelta(hours=1)))
        self.check_ongoing_count(date(2021,7,6), 0)
        self.check_ongoing_count(date(2021,7,7), 0)

        # With an endtime, crossing midnight
        Calendar.new_entry(EntryInfo(desc='with timezone, crossing midnight', start_dt=datetime(2021,7,8,12,0,tzinfo=tz_LON), duration=timedelta(hours=24)))
        self.check_ongoing_count(date(2021,7,8), 0)
        self.check_ongoing_count(date(2021,7,9), 1)
        self.check_ongoing_count(date(2021,7,10), 0)

        # With an endtime, crossing midnight locally
        tz_SAM = tz.gettz('Pacific/Samoa') # UTC-11
        self.assertTrue(tz_SAM) # check not None
        Calendar.new_entry(EntryInfo(desc='with westerly non-local timezone & endtime', start_dt=datetime(2021,7,10,0,5,tzinfo=tz_SAM), duration=timedelta(hours=23, minutes=50)))
        self.check_ongoing_count(date(2021,7,10), 0)
        self.check_ongoing_count(date(2021,7,11), 1)
        self.check_ongoing_count(date(2021,7,12), 0)

        # We want to test somewhere that is ahead of local time, so
        # let's take the most easterly timezone we know of.
        tz_NZ = tz.gettz('Pacific/Auckland')
        self.assertTrue(tz_NZ) # check not None

        # Check to see if local timezone is same as event timezone,
        # and if so change local tz to something "neutral".
        if get_local_tz().utcoffset(datetime(2021,7,12,12))>=timedelta(hours=12):
            set_local_tz(tz_LON)

        Calendar.new_entry(EntryInfo(desc='with easterly non-local timezone & endtime', start_dt=datetime(2021,7,13,0,5,tzinfo=tz_NZ), duration=timedelta(hours=23, minutes=50)))
        self.check_ongoing_count(date(2021,7,12), 0)
        self.check_ongoing_count(date(2021,7,13), 1)
        self.check_ongoing_count(date(2021,7,14), 0)


    #@unittest.skip
    def test_08_new_year(self) -> None:
        # Test cases over new year

        Calendar.new_entry(EntryInfo(desc='Stopping new year 2024', start_dt=datetime(2023,12,31,11,58), end_dt=datetime(2024,1,1,0,0)))
        self.check_ongoing_count(date(2023,12,31), 0)
        self.check_ongoing_count(date(2024,1,1), 0)
        self.check_ongoing_count(date(2024,1,2), 0)

        Calendar.new_entry(EntryInfo(desc='Starting new year 2024', start_dt=datetime(2024,1,1,0,0), end_dt=datetime(2024,1,1,23,0)))
        self.check_ongoing_count(date(2023,12,31), 0)
        self.check_ongoing_count(date(2024,1,1), 0)
        self.check_ongoing_count(date(2024,1,2), 0)

        Calendar.new_entry(EntryInfo(desc='Over new year 2024', start_dt=datetime(2023,12,31,11,58), end_dt=datetime(2024,1,1,0,1)))
        self.check_ongoing_count(date(2023,12,31), 0)
        self.check_ongoing_count(date(2024,1,1), 1)
        self.check_ongoing_count(date(2024,1,2), 0)

        Calendar.new_entry(EntryInfo(desc='Stopping new year 2023 (duration)', start_dt=datetime(2022,12,31,11,11), duration=timedelta(hours=12,minutes=49)))
        self.check_ongoing_count(date(2022,12,31), 0)
        self.check_ongoing_count(date(2023,1,1), 0)
        self.check_ongoing_count(date(2023,1,2), 0)

        Calendar.new_entry(EntryInfo(desc='Over new year 2023 (duration)', start_dt=datetime(2022,12,31,11,11), duration=timedelta(hours=12,minutes=50)))
        self.check_ongoing_count(date(2022,12,31), 0)
        self.check_ongoing_count(date(2023,1,1), 1)
        self.check_ongoing_count(date(2023,1,2), 0)

        # day entries
        Calendar.new_entry(EntryInfo(desc='Stopping new year 2022', start_dt=date(2021,12,30), end_dt=date(2022,1,1)))
        self.check_ongoing_count(date(2021,12,30), 0)
        self.check_ongoing_count(date(2021,12,31), 1)
        self.check_ongoing_count(date(2022,1,1), 0)
        self.check_ongoing_count(date(2022,1,2), 0)

        Calendar.new_entry(EntryInfo(desc='Starting new year 2022', start_dt=date(2022,1,1), end_dt=date(2022,1,2)))
        self.check_ongoing_count(date(2021,12,31), 1)
        self.check_ongoing_count(date(2022,1,1), 0)
        self.check_ongoing_count(date(2022,1,2), 0)

        Calendar.new_entry(EntryInfo(desc='Over new year 2022', start_dt=date(2021,12,31), end_dt=date(2022,1,2)))
        self.check_ongoing_count(date(2021,12,31), 1)
        self.check_ongoing_count(date(2022,1,1), 1)
        self.check_ongoing_count(date(2022,1,2), 0)

        Calendar.new_entry(EntryInfo(desc='Week-long over new year 2022', start_dt=date(2021,12,29), end_dt=date(2022,1,5)))
        self.check_ongoing_count(date(2021,12,29), 0)
        self.check_ongoing_count(date(2021,12,30), 1)
        self.check_ongoing_count(date(2021,12,31), 2)
        self.check_ongoing_count(date(2022,1,1), 2)
        self.check_ongoing_count(date(2022,1,2), 1)
        self.check_ongoing_count(date(2022,1,3), 1)
        self.check_ongoing_count(date(2022,1,4), 1)
        self.check_ongoing_count(date(2022,1,5), 0)

        # Over multiple years
        Calendar.new_entry(EntryInfo(desc='Event over new years 2020 & 2021', start_dt=date(2019,11,12), end_dt=date(2021,3,1)))
        self.check_ongoing_count(date(2019,11,12), 0)
        self.check_ongoing_count(date(2019,11,13), 1)
        self.check_ongoing_count(date(2019,11,30), 1)
        self.check_ongoing_count(date(2019,12,1), 1)
        self.check_ongoing_count(date(2019,12,31), 1)
        self.check_ongoing_count(date(2020,1,1), 1)
        self.check_ongoing_count(date(2020,1,31), 1)
        self.check_ongoing_count(date(2020,2,1), 1)
        self.check_ongoing_count(date(2020,2,28), 1)
        self.check_ongoing_count(date(2020,2,29), 1) # leap day
        self.check_ongoing_count(date(2020,3,1), 1)
        self.check_ongoing_count(date(2020,7,18), 1)
        self.check_ongoing_count(date(2020,11,12), 1)
        self.check_ongoing_count(date(2020,11,13), 1)
        self.check_ongoing_count(date(2020,11,30), 1)
        self.check_ongoing_count(date(2020,12,1), 1)
        self.check_ongoing_count(date(2020,12,31), 1)
        self.check_ongoing_count(date(2021,1,1), 1)
        self.check_ongoing_count(date(2021,1,31), 1)
        self.check_ongoing_count(date(2021,2,1), 1)
        self.check_ongoing_count(date(2021,2,28), 1)
        self.check_ongoing_count(date(2021,3,1), 0)
        self.check_ongoing_count(date(2021,3,2), 0)
        self.check_ongoing_count(date(2021,3,31), 0)


    #@unittest.skip
    def test_09_daylight_saving_start(self) -> None:
        # Test cases with daylight saving (so also has timezone)

        tz_BER = tz.gettz('Europe/Berlin')
        self.assertTrue(tz_BER) # check not None
        set_local_tz(tz_BER)

        Calendar.new_entry(EntryInfo(desc='over DST start, within day', start_dt=datetime(2024,3,31,0,30,tzinfo=tz_BER), end_dt=datetime(2024,3,31,23,55,tzinfo=tz_BER)))
        self.check_ongoing_count(date(2024,3,31), 0)
        self.check_ongoing_count(date(2024,4,1), 0)

        Calendar.new_entry(EntryInfo(desc='over DST start, less than day', start_dt=datetime(2024,3,31,0,30,tzinfo=tz_BER), end_dt=datetime(2024,4,1,0,1,tzinfo=tz_BER)))
        self.check_ongoing_count(date(2024,3,31), 0)
        self.check_ongoing_count(date(2024,4,1), 1)
        self.check_ongoing_count(date(2024,4,2), 0)

        Calendar.new_entry(EntryInfo(desc='over DST start, duration within day', start_dt=datetime(2023,3,26,0,30,tzinfo=tz_BER), duration=timedelta(hours=22)))
        self.check_ongoing_count(date(2023,3,26), 0)
        self.check_ongoing_count(date(2023,3,27), 0)

        Calendar.new_entry(EntryInfo(desc='over DST start, duration over day', start_dt=datetime(2023,3,26,0,30,tzinfo=tz_BER), duration=timedelta(hours=23)))
        self.check_ongoing_count(date(2023,3,26), 0)
        self.check_ongoing_count(date(2023,3,27), 1)
        self.check_ongoing_count(date(2023,3,28), 0)

        Calendar.new_entry(EntryInfo(desc='over DST start, day event', start_dt=date(2022,3,27), end_dt=date(2022,3,28)))
        self.check_ongoing_count(date(2022,3,27), 0)
        self.check_ongoing_count(date(2022,3,28), 0)

        Calendar.new_entry(EntryInfo(desc='over DST start, 2-day event', start_dt=date(2022,3,27), end_dt=date(2022,3,29)))
        self.check_ongoing_count(date(2022,3,27), 0)
        self.check_ongoing_count(date(2022,3,28), 1)
        self.check_ongoing_count(date(2022,3,29), 0)

        Calendar.new_entry(EntryInfo(desc='over DST start, 4-day event', start_dt=date(2022,3,26), end_dt=date(2022,3,30)))
        self.check_ongoing_count(date(2022,3,26), 0)
        self.check_ongoing_count(date(2022,3,27), 1)
        self.check_ongoing_count(date(2022,3,28), 2)
        self.check_ongoing_count(date(2022,3,29), 1)
        self.check_ongoing_count(date(2022,3,30), 0)

        Calendar.new_entry(EntryInfo(desc='over DST start, day event from duration', start_dt=date(2021,3,27), duration=timedelta(days=1)))
        self.check_ongoing_count(date(2021,3,27), 0)
        self.check_ongoing_count(date(2021,3,28), 0)

        Calendar.new_entry(EntryInfo(desc='over DST start, 2-day event from duration', start_dt=date(2021,3,27), duration=timedelta(days=2)))
        self.check_ongoing_count(date(2021,3,27), 0)
        self.check_ongoing_count(date(2021,3,28), 1)
        self.check_ongoing_count(date(2021,3,29), 0)

        Calendar.new_entry(EntryInfo(desc='over DST start, 4-day event from duration', start_dt=date(2021,3,26), duration=timedelta(days=4)))
        self.check_ongoing_count(date(2021,3,26), 0)
        self.check_ongoing_count(date(2021,3,27), 1)
        self.check_ongoing_count(date(2021,3,28), 2)
        self.check_ongoing_count(date(2021,3,29), 1)
        self.check_ongoing_count(date(2021,3,30), 0)


    #@unittest.skip
    def test_10_daylight_saving_end(self) -> None:
        # Test cases with daylight saving (so also has timezone)

        tz_CHI = tz.gettz('America/Chicago')
        self.assertTrue(tz_CHI) # check not None
        set_local_tz(tz_CHI)

        Calendar.new_entry(EntryInfo(desc='over DST end, within day', start_dt=datetime(2023,11,5,0,30,tzinfo=tz_CHI), end_dt=datetime(2023,11,5,23,55,tzinfo=tz_CHI)))
        self.check_ongoing_count(date(2023,11,5), 0)
        self.check_ongoing_count(date(2023,11,6), 0)

        Calendar.new_entry(EntryInfo(desc='over DST end, duration within day', start_dt=datetime(2022,11,6,0,30,tzinfo=tz_CHI), duration=timedelta(hours=23, minutes=59)))
        self.check_ongoing_count(date(2022,11,6), 0)
        self.check_ongoing_count(date(2022,11,7), 0)

        Calendar.new_entry(EntryInfo(desc='over DST end, duration one day', start_dt=datetime(2022,11,6,0,30,tzinfo=tz_CHI), duration=timedelta(days=1)))
        self.check_ongoing_count(date(2022,11,6), 0)
        self.check_ongoing_count(date(2022,11,7), 1)
        self.check_ongoing_count(date(2022,11,8), 0)

        Calendar.new_entry(EntryInfo(desc='over DST end, day event', start_dt=date(2021,11,7), end_dt=date(2021,11,8)))
        self.check_ongoing_count(date(2021,11,7), 0)
        self.check_ongoing_count(date(2021,11,8), 0)

        Calendar.new_entry(EntryInfo(desc='over DST end, 2-day event', start_dt=date(2021,11,7), end_dt=date(2021,11,9)))
        self.check_ongoing_count(date(2021,11,7), 0)
        self.check_ongoing_count(date(2021,11,8), 1)
        self.check_ongoing_count(date(2021,11,9), 0)

        Calendar.new_entry(EntryInfo(desc='over DST end, 4-day event', start_dt=date(2021,11,6), end_dt=date(2021,11,10)))
        self.check_ongoing_count(date(2021,11,6), 0)
        self.check_ongoing_count(date(2021,11,7), 1)
        self.check_ongoing_count(date(2021,11,8), 2)
        self.check_ongoing_count(date(2021,11,9), 1)
        self.check_ongoing_count(date(2021,11,10), 0)

        Calendar.new_entry(EntryInfo(desc='over DST end, day event from duration', start_dt=date(2020,11,1), duration=timedelta(days=1)))
        self.check_ongoing_count(date(2020,11,1), 0)
        self.check_ongoing_count(date(2020,11,2), 0)

        Calendar.new_entry(EntryInfo(desc='over DST end, 2-day event from duration', start_dt=date(2020,11,1), duration=timedelta(days=2)))
        self.check_ongoing_count(date(2020,11,1), 0)
        self.check_ongoing_count(date(2020,11,2), 1)
        self.check_ongoing_count(date(2020,11,3), 0)

        Calendar.new_entry(EntryInfo(desc='over DST end, 4-day event from duration', start_dt=date(2020,10,31), duration=timedelta(days=4)))
        self.check_ongoing_count(date(2020,10,31), 0)
        self.check_ongoing_count(date(2020,11,1), 1)
        self.check_ongoing_count(date(2020,11,2), 2)
        self.check_ongoing_count(date(2020,11,3), 1)
        self.check_ongoing_count(date(2020,11,4), 0)


    def check_ongoing_count(self, dt:date, count:int) -> None:
        # Helper function checks number of ongoing entries
        ol = Calendar.ongoing_list(dt)
        #print(ol)
        self.assertEqual(len(ol), count)


# Run all tests if this file is executed as main
if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# pygenda_view_todo.py
# Provides the "To-Do View" for Pygenda.
#
# Copyright (C) 2022-2024 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from gi.repository import Gtk, Gdk, GLib
from gi.repository.Pango import WrapMode as PWrapMode

from icalendar import cal as iCal, Event as iEvent, Todo as iTodo
from locale import gettext as _
from typing import Optional, List, Tuple, Union

# pygenda components
from .pygenda_view import View
from .pygenda_calendar import Calendar
from .pygenda_config import Config
from .pygenda_gui import GUI
from .pygenda_dialog_todo import TodoDialogController
from .pygenda_entryinfo import EntryInfo


# Singleton class for Todo View
class View_Todo(View):
    Config.set_defaults('todo_view',{
        'list0_title': _('To-do'),
        'zoom_levels': 5,
        'default_zoom': 2,
    })

    _cursor_list = 0
    _cursor_idx_in_list = 0
    _last_cursor_list = None
    _last_cursor_idx_in_list = None
    _list_items = None # type: list
    _target_listidx = None
    _target_todo = None
    _target_cursor_y = None # type: Optional[float]
    _scroll_to_cursor_required = False
    CURSOR_STYLE = 'todoview_cursor'

    @staticmethod
    def view_name() -> str:
        # Return (localised) string to use in menu
        return _('_Todo View')

    @staticmethod
    def accel_key() -> int:
        # Return (localised) keycode for menu shortcut
        k = _('todo_view_accel')
        return ord(k[0]) if len(k)>0 else 0


    @classmethod
    def init(cls) -> Gtk.Widget:
        # Called on startup.
        # Returns widget containing view.
        cls._init_parse_list_config()
        cls._init_todo_widgets()
        cls._init_keymap()
        cls.init_zoom('todo_view', cls._topboxscroll.get_style_context())
        return cls._topboxscroll


    @classmethod
    def _init_parse_list_config(cls) -> None:
        # Read & parse config settings
        i = 0
        cls._list_titles = []
        cls._list_filters = []
        cls._list_default_cats = []
        while True:
            title = Config.get('todo_view','list{}_title'.format(i))
            if title is None:
                break
            filt = Config.get('todo_view','list{}_filter'.format(i))
            cls._list_titles.append(title)
            cls._list_filters.append(filt)
            cls._list_default_cats.append(cls._default_cats_from_filter(filt))
            i += 1
        cls._list_count = i
        cls._item_counts = [0]*cls._list_count


    @staticmethod
    def _default_cats_from_filter(filt:Optional[str]) -> Optional[list]:
        # Return list of categories that will match given filter string.
        # For now this is quite simple, but in the future it may support
        # more complex filters (e.g. project1 AND this_week).
        if not filt or filt=='UNCATEGORIZED':
            return None
        return [filt]


    @classmethod
    def _init_todo_widgets(cls) -> None:
        # Initialise widgets - create list labels, entry spaces etc.
        # First make the top-level container
        cls._topboxscroll = Gtk.ScrolledWindow()
        cls._topboxscroll.set_name('view_todo')
        cls._topboxscroll.set_policy(Gtk.PolicyType.AUTOMATIC,Gtk.PolicyType.NEVER)
        cls._topboxscroll.set_overlay_scrolling(False)
        cls._topboxscroll.set_hexpand(True)
        list_hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        list_hbox.set_homogeneous(True)
        cls._topboxscroll.add(list_hbox)
        list_hbox.connect('draw', cls._pre_draw)

        # Now add vertical boxes for each list
        cls._list_box = []
        cls._list_scroller = []
        for i in range(cls._list_count):
            new_list = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
            new_list.get_style_context().add_class('todoview_list')
            new_title = Gtk.Label(cls._list_titles[i])
            new_title.set_line_wrap(True)
            new_title.set_line_wrap_mode(PWrapMode.WORD_CHAR)
            new_title.get_style_context().add_class('todoview_title')
            new_eventbox_title = Gtk.EventBox()
            new_eventbox_title.connect('button_press_event', cls.click_title, i)
            new_eventbox_title.add(new_title)
            new_list.add(new_eventbox_title)
            new_list_scroller = Gtk.ScrolledWindow()
            new_list_scroller.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
            new_list_scroller.set_overlay_scrolling(False)
            new_list_scroller.set_vexpand(True)
            new_list_scroller.get_style_context().add_class('todoview_listcontent')
            new_list_scroller.get_vadjustment().connect('value-changed', cls._list_scrolled)
            new_eventbox_list = Gtk.EventBox()
            new_eventbox_list.connect('button_press_event', cls.click_list, i)
            new_eventbox_list.add(new_list_scroller)
            new_list.add(new_eventbox_list)
            new_list_content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
            new_list_scroller.add(new_list_content)
            cls._list_box.append(new_list)
            cls._list_scroller.append(new_list_scroller)
            list_hbox.add(new_list)


    @classmethod
    def _init_keymap(cls) -> None:
        # Initialises KEYMAP for class. Called from init() since it needs
        # to be called after class construction, so that functions exist.
        cls._KEYMAP = {
            Gdk.KEY_Up: lambda: cls._cursor_move_up(),
            Gdk.KEY_Down: lambda: cls._cursor_move_dn(),
            Gdk.KEY_Right: lambda: cls._cursor_move_rt(),
            Gdk.KEY_Left: lambda: cls._cursor_move_lt(),
            Gdk.KEY_Home: lambda: cls._cursor_move_list(0),
            Gdk.KEY_End: lambda: cls._cursor_move_list(-1),
            Gdk.KEY_Page_Up: lambda: cls._cursor_move_index(0),
            Gdk.KEY_Page_Down: lambda: cls._cursor_move_index(-1),
            Gdk.KEY_Return: lambda: cls.cursor_edit_entry(),
        }


    @classmethod
    def get_cursor_entry(cls) -> Optional[iTodo]:
        # Returns entry at cursor position, or None if cursor not on entry.
        # Called from cursor_edit_entry() & delete_request().
        if len(cls._list_items[cls._cursor_list]) == 0:
            return None
        return cls._list_items[cls._cursor_list][cls._cursor_idx_in_list]


    @classmethod
    def default_entry_is_todo(cls) -> bool:
        # Returns True: creating an "entry" in Todo View -> new Todo
        return True


    @classmethod
    def paste_entry(cls, en:Union[iEvent,iTodo]) -> None:
        # Creates new entry based on entry en.
        # Type of entry depends on View (e.g. Todo View -> to-do item).
        cats = cls._list_default_cats[cls._cursor_list]
        new_en = Calendar.paste_entry(en, e_type=EntryInfo.TYPE_TODO, e_cats=cats)
        cls.cursor_goto_todo(new_en, cls._cursor_list)


    @classmethod
    def paste_text(cls, txt:str) -> None:
        # Handle pasting of text in Todo view.
        # Open a New Todo dialog with description initialised as txt,
        # and to-do list set from current cursor position.
        GLib.idle_add(lambda x: TodoDialogController.new_todo(txt=x, list_idx=cls.cursor_todo_list()), txt)


    @classmethod
    def cursor_todo_list(cls) -> int:
        # Returns index of todo list with cursor.
        # Used by "new todo" etc to initialise dialog with focused todo list.
        return cls._cursor_list


    @classmethod
    def cursor_goto_todo(cls, todo:iTodo, list_idx:int) -> bool:
        # Move cursor to given todo in given list.
        # Return True to indicate this is possible in this view.
        # Set targets, but don't move yet, since we want to redraw view
        # & to use recalculated todo order to determine cursor position.
        cls._target_listidx = list_idx
        cls._target_todo = todo
        return True


    @classmethod
    def redraw(cls, en_changes:bool) -> None:
        # Called when redraw required
        # en_changes: bool indicating if displayed entries need updating too
        cls._target_cursor_y = None # reset navigation y-coord
        if not en_changes:
            return
        cls._last_cursor_list = None
        cls._last_cursor_idx_in_list = None
        for cont in cls._list_scroller:
            cont.get_child().destroy()
        todos = Calendar.todo_list()
        cls._list_items = []
        for i in range(len(cls._list_scroller)):
            new_list_content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
            count = 0
            cls._list_items.append([])
            for td in todos:
                if cls._todo_matches_filter(td, cls._list_filters[i]):
                    if cls._target_todo is not None and cls._target_todo is td:
                        cls._cursor_list = i
                        cls._cursor_idx_in_list = count
                        if i>=cls._target_listidx:
                            # We've reached the target list, so go no further
                            cls._target_listidx = None
                            cls._target_todo = None
                    row = Gtk.Box()
                    ctx = row.get_style_context()
                    ctx.add_class('todoview_item')
                    # Potential markers: ①-0x245f ➀-0x277f ❶-0x2775 ➊-0x2789
                    mark_label = Gtk.Label(chr(0x2789+td['PRIORITY']) if 'PRIORITY' in td else u'•')
                    mark_label.set_halign(Gtk.Align.END)
                    mark_label.set_valign(Gtk.Align.START)
                    mark_label.get_style_context().add_class('todoview_marker')
                    row.add(mark_label)
                    txt = ''
                    if 'DUE' in td:
                        due_dt = td['DUE'].dt
                        due_dt_st = due_dt.strftime(GUI.date_formatting_numeric)
                        txt += '({:s} {:s}) '.format(_('Due:'), due_dt_st)
                    txt += td['SUMMARY'] if 'SUMMARY' in td else ''
                    txt += cls.entry_icons(td, True)
                    item_text = Gtk.Label(txt)
                    item_text.get_style_context().add_class('todoview_itemtext')
                    item_text.set_xalign(0)
                    item_text.set_yalign(0)
                    item_text.set_line_wrap(True)
                    item_text.set_line_wrap_mode(PWrapMode.WORD_CHAR)
                    if 'STATUS' in td and td['STATUS'] in Calendar.STATUS_LIST_TODO:
                        ctx.add_class(td['STATUS'].lower())
                    row.add(item_text)
                    new_list_content.add(row)
                    cls._list_items[-1].append(td)
                    count += 1
            cls._item_counts[i] = count
            if count==0:
                # an empty list, need something for cursor
                mark_label = Gtk.Label()
                mark_label.set_halign(Gtk.Align.START) # else cursor fills line
                ctx = mark_label.get_style_context()
                ctx.add_class('todoview_marker')
                new_list_content.add(mark_label)
            new_list_content.get_style_context().add_class('todoview_items')
            cls._list_scroller[i].add(new_list_content)
            cls._list_scroller[i].show_all()
        # Reset target (though in theory this should already be done)
        cls._target_listidx = None
        cls._target_todo = None
        cls._show_cursor()


    @classmethod
    def zoom(cls, inc:int) -> None:
        # Override zoom() in parent class so we can reset y-coord target
        super().zoom(inc) # call version of zoom() in parent class
        cls._target_cursor_y = None


    @staticmethod
    def _todo_matches_filter(td:iTodo, filt:Optional[str]) -> bool:
        # Return True if categories of to-do item match filter
        if filt is None:
            return True
        cats = View_Todo._get_categories(td)
        if not cats:
            return filt=='UNCATEGORIZED'
        return filt in cats


    @staticmethod
    def first_list_index(td:iTodo) -> Optional[int]:
        # Return index of first todo list containing td
        for i in range(len(View_Todo._list_filters)):
            if View_Todo._todo_matches_filter(td, View_Todo._list_filters[i]):
                return i
        return None


    @staticmethod
    def _get_categories(td:iTodo) -> list:
        # Return list of categories of the given to-do item
        if 'CATEGORIES' not in td:
            cats = [] # type: List[str]
        elif isinstance(td['CATEGORIES'],list):
            cats = []
            for clist in td['CATEGORIES']:
                if isinstance(clist,str):
                    cats.extend([c for c in clist.split(',') if c])
                else:
                    cats.extend([c for c in clist.cats if c])
        elif isinstance(td['CATEGORIES'],str):
            cats = [c for c in td['CATEGORIES'].split(',') if c]
        else:
            cats = [c for c in td['CATEGORIES'].cats if c]
        return cats


    @classmethod
    def click_title(cls, wid:Gtk.Widget, ev:Gdk.EventButton, list_idx:int) -> bool:
        # Callback. Called whenever a list title is clicked/tapped.
        # Moves cursor to top of list/item clicked
        cls._cursor_move_list(list_idx)
        cls._cursor_move_index(0)
        cls._target_cursor_y = None
        cls._scroll_to_cursor_required = True
        return True # event handled - don't propagate


    @classmethod
    def click_list(cls, wid:Gtk.Widget, ev:Gdk.EventButton, list_idx:int) -> bool:
        # Callback. Called whenever list content is clicked/tapped.
        # Moves cursor to list/item clicked
        cls._cursor_move_list(list_idx)
        scroller = cls._list_scroller[list_idx]
        vwport = scroller.get_children()[0]
        row_vbox = vwport.get_children()[0]
        # Pixels above the first row:
        ycount = cls._top_spacing(scroller)
        ycount += cls._top_spacing(vwport)
        ycount += cls._top_spacing(row_vbox)
        # Take account of list's vertical scrollbar:
        ycount -= cls._list_scroller[list_idx].get_vadjustment().get_value()
        # Look through rows until cumulative height exceeds click ycoord
        row_spacing = row_vbox.get_spacing()
        rows = row_vbox.get_children()
        ycount -= row_spacing//2 # adjust for no spacing above first row
        i = 0
        for r in rows:
            ycount += r.get_allocated_height()
            ycount += row_spacing
            if ycount > ev.y:
                break
            i += 1
        cls._cursor_move_index(i)
        cls._target_cursor_y = None
        cls._scroll_to_cursor_required = True
        return True # event handled - don't propagate


    @classmethod
    def _list_scrolled(cls, wid:Gtk.Widget) -> bool:
        # Callback. Called whenever a list is scrolled
        cls._target_cursor_y = None # reset navigation y-coord target
        return True # don't propagate event


    @staticmethod
    def _top_spacing(wid:Gtk.Widget) -> float:
        # Return total size of top padding+border+margin of widget
        ctx = wid.get_style_context()
        pad = ctx.get_padding(Gtk.StateFlags.NORMAL)
        bord = ctx.get_border(Gtk.StateFlags.NORMAL)
        marg = ctx.get_margin(Gtk.StateFlags.NORMAL)
        return pad.top + bord.top + marg.top # type:ignore


    @staticmethod
    def _vert_spacing(wid:Gtk.Widget) -> float:
        # Return total size of top+bottom padding+border+margin of widget
        ctx = wid.get_style_context()
        pad = ctx.get_padding(Gtk.StateFlags.NORMAL)
        bord = ctx.get_border(Gtk.StateFlags.NORMAL)
        marg = ctx.get_margin(Gtk.StateFlags.NORMAL)
        return pad.top+pad.bottom + bord.top+bord.bottom + marg.top+marg.bottom # type:ignore


    @staticmethod
    def _left_spacing(wid:Gtk.Widget) -> float:
        # Return total size of left padding+border+margin of widget
        ctx = wid.get_style_context()
        pad = ctx.get_padding(Gtk.StateFlags.NORMAL)
        bord = ctx.get_border(Gtk.StateFlags.NORMAL)
        marg = ctx.get_margin(Gtk.StateFlags.NORMAL)
        return pad.left + bord.left + marg.left # type:ignore


    @staticmethod
    def _side_spacing(wid:Gtk.Widget) -> float:
        # Return total size of left+right padding+border+margin of widget
        ctx = wid.get_style_context()
        pad = ctx.get_padding(Gtk.StateFlags.NORMAL)
        bord = ctx.get_border(Gtk.StateFlags.NORMAL)
        marg = ctx.get_margin(Gtk.StateFlags.NORMAL)
        return pad.left+pad.right + bord.left+bord.right + marg.left+marg.right # type:ignore


    @classmethod
    def _show_cursor(cls) -> None:
        # Locates bullet corresponding to the current cursor and adds
        # 'todoview_cursor' class to it, so cursor is visible via CSS styling.

        # First correct cursor if required (e.g. item was deleted)
        if not (0 <= cls._cursor_list < cls._list_count):
            cls._cursor_list = max(0, cls._list_count-1)
        icount = cls._item_counts[cls._cursor_list]
        if not (0 <= cls._cursor_idx_in_list < icount):
            cls._cursor_idx_in_list = max(0, icount-1)

        cls._hide_cursor()

        ctx = cls._get_cursor_ctx(cls._cursor_list, cls._cursor_idx_in_list)
        if ctx is not None:
            ctx.add_class(cls.CURSOR_STYLE)
        cls._last_cursor_list = cls._cursor_list
        cls._last_cursor_idx_in_list = cls._cursor_idx_in_list

        # Enable/disable menu items
        ot = (icount!=0)
        ro = ot and Calendar.calendar_readonly(cls.get_cursor_entry())
        cch = GUI.create_todos
        GUI.set_menu_elts(on_todo=ot, read_only=ro, can_create_here=cch)

        # Add scroll call to idle, in case column widths not yet determined
        cls._scroll_to_cursor_required = True

        # If cursor is offscreen, need signal that a redraw is needed
        cls._topboxscroll.get_child().queue_draw()


    @classmethod
    def _pre_draw(cls, wid:Gtk.Widget, _) -> bool:
        # Callback called on 'draw' event on date_content.
        # Called before drawing date content.
        # Used to scroll window when cursor has been moved (since we
        # need to have calculated the layout to know where to scoll to).
        if cls._scroll_to_cursor_required:
            cls._scroll_to_cursor()
            cls._scroll_to_cursor_required = False
        return False # propagate event


    @classmethod
    def _hide_cursor(cls) -> None:
        # Clears 'todoview_cursor' style class from cursor position,
        # so cursor is no longer visible.
        if cls._last_cursor_list is not None:
            ctx = cls._get_cursor_ctx(cls._last_cursor_list, cls._last_cursor_idx_in_list)
            if ctx is not None:
                ctx.remove_class(cls.CURSOR_STYLE)
            cls._last_cursor_list = None
            cls._last_cursor_idx_in_list = None


    @classmethod
    def _get_cursor_ctx(cls, c_list:int, c_i_in_list:int) -> Gtk.StyleContext:
        # Returns a StyleContext object for to-do cursor coordinates
        # c_list & c_i_in_list
        lst = cls._list_scroller[c_list].get_child().get_child()
        item = lst.get_children()[c_i_in_list]
        if cls._item_counts[c_list]==0:
            ci = item
        else:
            ci = item.get_children()[0]
        return ci.get_style_context()


    @classmethod
    def _scroll_to_cursor(cls) -> None:
        # If required, scroll view elements so that cursor is visible.
        # Note: If view is not yet laid-out, calculation not correct.
        # Hence this will be called in 'draw' event handler, so layout is done.

        # First: horizontal scrolling...
        # (We know spacing for hbox is zero, since we control it, so ignore)
        list_width = cls._list_box[0].get_allocated_width() # homogeneous
        viewport = cls._topboxscroll.get_child()
        view_width = viewport.get_allocated_width()
        maxv = cls._cursor_list*list_width # left edge of list at left of view
        maxv += cls._left_spacing(viewport.get_child())
        minv = maxv + list_width - view_width # rt edge @ rt of view
        minv += cls._side_spacing(viewport)
        adj = cls._topboxscroll.get_hadjustment()
        cur = adj.get_value()
        if minv > maxv:
            minv = maxv # If list is wider than view, then show left edge
        if cur > maxv:
            adj.set_value(maxv)
        elif cur < minv:
            adj.set_value(minv)

        # Now the vertical scrolling...
        list_scroller, listitems_box = cls._current_scroller_itemsbox()
        viewport = list_scroller.get_child()
        # Take into account padding/margin etc of todoview_items:
        maxv = cls._top_spacing(listitems_box)
        maxv += listitems_box.get_spacing() * cls._cursor_idx_in_list
        list_item_wids = listitems_box.get_children()
        for i in range(cls._cursor_idx_in_list):
            maxv += list_item_wids[i].get_allocated_height()
        minv = maxv + list_item_wids[cls._cursor_idx_in_list].get_allocated_height()
        # For min, subtract visible content height (child in case of padding):
        minv -= viewport.get_allocated_height()
        minv += cls._vert_spacing(viewport)
        if minv > maxv:
            minv = maxv # If item is taller than view, then show top
        adj = list_scroller.get_vadjustment()
        cur = adj.get_value()
        if cur > maxv:
            adj.set_value(maxv)
        elif cur < minv:
            adj.set_value(minv)


    @classmethod
    def keypress(cls, wid:Gtk.Widget, ev:Gdk.EventKey) -> None:
        # Handle key press event in Todo view.
        # Called (from GUI.keypress()) on keypress (or repeat) event
        try:
            f = cls._KEYMAP[ev.keyval]
            GLib.idle_add(f)
        except KeyError:
            # If it's a character key, take as first of new todo
            # !! Bug: only works for ASCII characters
            if GUI.create_todos and ev.state & (Gdk.ModifierType.CONTROL_MASK|Gdk.ModifierType.MOD1_MASK)==0 and Gdk.KEY_exclam <= ev.keyval <= Gdk.KEY_asciitilde:
                GLib.idle_add(lambda x: TodoDialogController.new_todo(txt=x, list_idx=cls.cursor_todo_list()), chr(ev.keyval))


    @classmethod
    def _cursor_move_up(cls) -> None:
        # Callback for user moving cursor up
        cls._cursor_idx_in_list -= 1 # Cursor correction will fix if <0
        cls._target_cursor_y = None
        cls._show_cursor()

    @classmethod
    def _cursor_move_dn(cls) -> None:
        # Callback for user moving cursor down
        cls._target_cursor_y = None
        if cls._item_counts[cls._cursor_list] > 0:
            cls._cursor_idx_in_list = (cls._cursor_idx_in_list+1)%cls._item_counts[cls._cursor_list]
            cls._show_cursor()

    @classmethod
    def _cursor_move_rt(cls) -> None:
        # Callback for user moving cursor right
        if cls._target_cursor_y is None:
            cls._target_cursor_y = cls._get_cursor_y()
        cls._cursor_list = (cls._cursor_list+1)%cls._list_count
        cls._cursor_idx_in_list = cls._y_to_todo_index(cls._target_cursor_y)
        cls._show_cursor()

    @classmethod
    def _cursor_move_lt(cls) -> None:
        # Callback for user moving cursor left
        if cls._target_cursor_y is None:
            cls._target_cursor_y = cls._get_cursor_y()
        cls._cursor_list -= 1 # Cursor correction will fix if <0
        cls._cursor_idx_in_list = cls._y_to_todo_index(cls._target_cursor_y)
        cls._show_cursor()

    @classmethod
    def _cursor_move_list(cls, lst:int) -> None:
        # Callback for user moving cursor to list
        cls._cursor_list = lst
        cls._show_cursor()

    @classmethod
    def _cursor_move_index(cls, idx:int) -> None:
        # Callback for user moving cursor to idx in current list
        cls._cursor_idx_in_list = idx
        cls._show_cursor()

    @classmethod
    def cursor_edit_entry(cls) -> None:
        # Opens a todo edit dialog for the entry at the cursor,
        # or to create a new todo if the cursor is not on entry.
        # Assigned to the 'Enter' key.
        en = cls.get_cursor_entry()
        if en is not None:
            GUI.edit_or_display_todo(en, list_idx=cls.cursor_todo_list())
        elif GUI.create_todos:
            TodoDialogController.new_todo(list_idx=cls.cursor_todo_list())


    @classmethod
    def _get_cursor_y(cls) -> float:
        # Return "visual" y of cursor - i.e. y coord as cursor
        # is displayed. Used to calculate which todo item to
        # move to when the cursor is moved right or left.
        y = cls._top_spacing(cls._list_box[cls._cursor_list])
        list_title = cls._current_list_title()
        y += list_title.get_allocated_height()
        list_scroller, listitems_box = cls._current_scroller_itemsbox()
        y += cls._top_spacing(list_scroller)
        y += cls._top_spacing(listitems_box)
        y += listitems_box.get_spacing() * cls._cursor_idx_in_list
        list_item_wids = listitems_box.get_children()
        for i in range(cls._cursor_idx_in_list):
            y += list_item_wids[i].get_allocated_height()
        y -= list_scroller.get_vadjustment().get_value() # Account for scrollbar
        return y # type: ignore


    @classmethod
    def _y_to_todo_index(cls, y:float) -> int:
        # Return index in current list, given a visual y-coord.
        # Used to calculate which todo item to move to when the
        # cursor is moved right or left.
        list_scroller, listitems_box = cls._current_scroller_itemsbox()
        y += list_scroller.get_vadjustment().get_value() # Account for scrollbar
        idx = 0
        list_title = cls._current_list_title()
        spacing = listitems_box.get_spacing()
        ycount = cls._top_spacing(cls._list_box[cls._cursor_list])
        ycount += list_title.get_allocated_height()
        ycount += cls._top_spacing(list_scroller)
        ycount += cls._top_spacing(listitems_box)
        for w in listitems_box.get_children():
            last_ycount = ycount
            ycount += w.get_allocated_height()
            ycount += spacing
            if ycount > y:
                if y-last_ycount > ycount-y:
                    # y is closer to next entry, so choose that one
                    idx += 1
                break
            idx += 1
        return idx


    @classmethod
    def _current_list_title(cls) -> Gtk.Label:
        # Utility function to return current list title Gtk.Label widget
        # ("current" meaning current cursor list).
        lst = cls._list_box[cls._cursor_list]
        tlab = lst.get_children()[0].get_child()
        return tlab


    @classmethod
    def _current_scroller_itemsbox(cls) -> Tuple[Gtk.ScrolledWindow,Gtk.Box]:
        # Utility function to return current list scroller and vbox
        # containing list items ("current" meaning current cursor list).
        list_scroller = cls._list_scroller[cls._cursor_list]
        listitems_box = list_scroller.get_child().get_child()
        return list_scroller, listitems_box



# -*- coding: utf-8 -*-
#
# pygenda_view_year.py
# Provides the "Year View" for Pygenda.
#
# Copyright (C) 2022-2024 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from gi.repository import Gtk, Gdk, GLib

import calendar
from datetime import date as dt_date, datetime as dt_datetime, timedelta
from locale import gettext as _
from icalendar import cal as iCal, Event as iEvent, Todo as iTodo
from typing import Tuple

# pygenda components
from .pygenda_view import View, View_DayUnit_Base
from .pygenda_gui import GUI
from .pygenda_dialog_event import EventDialogController
from .pygenda_config import Config
from .pygenda_calendar import Calendar
from .pygenda_util import start_end_dts_occ


# Singleton class for Year View
class View_Year(View_DayUnit_Base):
    Config.set_defaults('year_view',{
        'show_event_location': 'always',
        'show_todos': True,
        'zoom_levels': 5,
        'default_zoom': 2,
    })

    DAY_CLASS = [ 'yearview_day_{}'.format(s) for s in ['mon','tue','wed','thu','fri','sat','sun'] ]
    GRID_COLUMNS = 37
    GRID_ROWS = 12 # One per month
    GRID_CURSOR_STYLE = 'yearview_cursor'
    ENTRY_CURSOR_STYLE = 'yearview_entry_cursor'
    _target_col = None
    _year_viewed = -1 # Indicates next redraw will draw year
    _last_cursor = None
    _visible_occurrences = None
    _show_datecontent_pending = False
    _date_content_count = 0
    _scroll_to_cursor_required = False

    SHOW_LOC_ALWAYS = 1 # constant 'enum' for _show_location flag

    @staticmethod
    def view_name() -> str:
        # Return (localised) string to use in menu
        return _('_Year View')

    @staticmethod
    def accel_key() -> int:
        # Return (localised) keycode for menu shortcut
        k = _('year_view_accel')
        return ord(k[0]) if len(k)>0 else 0


    @classmethod
    def init(cls) -> Gtk.Widget:
        # Called on startup.
        # Gets view framework from UI file & tweaks/adds a few elements.
        # Returns widget containing view.
        GUI.load_ui_file('view_year.ui')
        cls._topbox = GUI._builder.get_object('view_year')
        cls._grid_cells = GUI._builder.get_object('year_grid_days')
        cls._date_label = GUI._builder.get_object('year_datelabel')
        cls._date_content_scroll = GUI._builder.get_object('year_datecontent_scroll')
        cls._date_content = GUI._builder.get_object('year_datecontent')
        cls._draw_day_month_labels()
        cls._init_keymap()
        cls._init_grid()
        cls._init_config()
        cls.init_zoom('year_view', cls._topbox.get_style_context())

        # Connect signal handlers
        HANDLERS = {
            'year_grid_click': cls.click_grid,
            'year_dtcont_click': cls.click_events,
            'year_dtcont_draw': cls._pre_datecontent_draw,
            }
        GUI._builder.connect_signals(HANDLERS)

        cls._init_gestures()

        return cls._topbox


    @classmethod
    def _draw_day_month_labels(cls) -> None:
        # Adds text to grid day and month labels.
        # Called only at initialisation.
        mon_names = GUI._builder.get_object('month_names')
        for i in range(cls.GRID_ROWS):
            l = Gtk.Label(calendar.month_abbr[i+1].capitalize())
            ctx = l.get_style_context()
            ctx.add_class('yearview_month_label')
            mon_names.add(l)

        st_wk = Config.get_int('global','start_week_day')
        day_names = GUI._builder.get_object('day_names')
        day_label_txt = [ d[0].capitalize() for d in calendar.day_abbr ]
        for i in range(st_wk,st_wk+cls.GRID_COLUMNS):
            l = Gtk.Label(day_label_txt[i%7])
            ctx = l.get_style_context()
            ctx.add_class('yearview_day_label')
            ctx.add_class(cls.DAY_CLASS[i%7])
            day_names.add(l)


    @classmethod
    def _init_keymap(cls) -> None:
        # Initialises KEYMAP for class. Called from init() since it needs
        # to be called after class construction, so that functions exist.
        cls._KEYMAP = {
            Gdk.KEY_Left: lambda: cls._cursor_move_lr(-1),
            Gdk.KEY_Right: lambda: cls._cursor_move_lr(1),
            Gdk.KEY_Up: lambda: cls._cursor_move_up(),
            Gdk.KEY_Down: lambda: cls._cursor_move_dn(),
            Gdk.KEY_Home: lambda: cls._cursor_move_stmon(),
            Gdk.KEY_End: lambda: cls._cursor_move_endmon(),
            Gdk.KEY_Page_Up: lambda: cls._cursor_move_pgupdn(-1),
            Gdk.KEY_Page_Down: lambda: cls._cursor_move_pgupdn(1),
            Gdk.KEY_space: lambda: cls._cursor_move_today(),
            Gdk.KEY_Return: lambda: cls.cursor_edit_entry(),
        }
        cls._KEYMAP_SHIFT = {
            Gdk.KEY_Up: lambda: cls._idxcursor_move_up(),
            Gdk.KEY_Down: lambda: cls._idxcursor_move_dn(),
            Gdk.KEY_Page_Up: lambda: cls._idxcursor_move_up(), # !! Placeholder
            Gdk.KEY_Page_Down: lambda: cls._idxcursor_move_dn(),#!! Placeholder
        }


    @classmethod
    def _init_grid(cls) -> None:
        # Adds labels to each grid cell
        for m in range(0,cls.GRID_ROWS):
            for d in range(0,cls.GRID_COLUMNS):
                l = Gtk.Label()
                l.set_xalign(0)
                l.set_yalign(0)
                cls._grid_cells.attach(l,d,m,1,1)


    @classmethod
    def _init_config(cls) -> None:
        # Initialisation from config settings.
        # Set _show_location flag from config.
        show_loc = Config.get('year_view','show_event_location')
        map = {'always':cls.SHOW_LOC_ALWAYS, 'never':0}
        cls._show_location = map[show_loc] if show_loc in map else 0
        cls._loc_max_chars = Config.get_int('year_view','location_max_chars')
        if cls._loc_max_chars is None:
            cls._loc_max_chars = 0
        cls.show_todos = Config.get_bool('year_view','show_todos')


    @classmethod
    def _init_gestures(cls) -> None:
        # Initialise swipe gesture. Called from init().
        yge = GUI._builder.get_object('year_grid_events')
        cls.gest_swipe = Gtk.GestureSwipe(widget=yge)
        cls.gest_swipe.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        cls.gest_swipe.set_touch_only(True)
        cls.gest_swipe.connect('swipe', cls._swipe_handler)


    @classmethod
    def _swipe_handler(cls, gest:Gtk.GestureSwipe, vx:float, vy:float) -> None:
        # Callback for swipe gestures (for navigation)
        VLIM = 70 # !! Magic number for now (?should use start/stop pos?)
        abs_vy = abs(vy)
        if abs_vy > VLIM and abs(vx)*2 < abs_vy:
            cls._cursor_move_pgupdn(1 if vy<0 else -1)
        # I don't think this type (GCallback) returns a value (unlike
        # other signal handlers, which return bools). So don't return anything.


    @classmethod
    def _draw_year(cls) -> None:
        # Draws year - adds formatting classes to cells, labels to some dates.
        # Called on redraw if year has changed.
        # Function a bit disorganised, tidying might be beneficial...
        c_date = View._cursor_date
        yr = c_date.year
        cls._year_viewed = yr
        l = GUI._builder.get_object('year_yearlabel')
        l.set_text(str(yr))

        st_wk = Config.get_int('global','start_week_day')
        mon_labs = GUI._builder.get_object('month_names').get_children()
        day_labs = GUI._builder.get_object('day_names').get_children()
        date = dt_date(year=yr,month=1,day=1)
        oneday = timedelta(days=1)
        today = dt_date.today()

        for m in range(1,cls.GRID_ROWS+1):
            day,daycount = calendar.monthrange(yr,m)
            col = (day-st_wk)%7 # Column for first of the month
            ctx = mon_labs[m-1].get_style_context()
            if col==0:
                ctx.add_class('yearview_leftofdaycell')
            else:
                ctx.remove_class('yearview_leftofdaycell')
            if m!=cls.GRID_ROWS:
                # This is to take account of grid lines above next month
                nday,ndaycount = calendar.monthrange(yr,m+1)
                ncol = (nday-st_wk)%7
                ndayend = ncol+ndaycount
            else:
                ncol = 7
                ndayend = 0
            for c in range(col): # Empty cells before col
                l = cls._grid_cells.get_child_at(c,m-1)
                l.set_text('')
                ctx = l.get_style_context()
                cls.remove_all_classes(ctx)
                ctx.add_class('yearview_emptycell')
                if c==col-1:
                    ctx.add_class('yearview_leftofdaycell')
                if c>=ncol:
                    ctx.add_class('yearview_abovedaycell')
                if m==1:
                    ctx = day_labs[c].get_style_context()
                    ctx.remove_class('yearview_abovedaycell')

            for d in range(daycount):
                # Potential optimisations here? Central columns are always in
                # year, so no need to remove&re-add class yearview_daycell etc.
                # Also if not start of week, will never be start of week,
                # so we know label text is already ''.
                t = ''
                if day==st_wk or d==0 or d==daycount-1:
                    t = str(d+1)
                l = cls._grid_cells.get_child_at(col,m-1)
                l.set_text(t)
                ctx = l.get_style_context()
                cls.remove_all_classes(ctx)
                ctx.add_class('yearview_daycell')
                ctx.add_class(cls.DAY_CLASS[day])
                if date<today:
                    ctx.add_class('yearview_pastday')
                elif date==today:
                    ctx.add_class('yearview_today')
                if m==1:
                    ctx = day_labs[col].get_style_context()
                    ctx.add_class('yearview_abovedaycell')
                day = (day+1)%7
                col += 1
                date += oneday

            for c in range(col,cls.GRID_COLUMNS): # Empty cells after
                l = cls._grid_cells.get_child_at(c,m-1)
                l.set_text('')
                ctx = l.get_style_context()
                cls.remove_all_classes(ctx)
                ctx.add_class('yearview_emptycell')
                if c<ndayend:
                    ctx.add_class('yearview_abovedaycell')
                if m==1:
                    ctx = day_labs[c].get_style_context()
                    ctx.remove_class('yearview_abovedaycell')


    @classmethod
    def renew_display(cls) -> None:
        # Called when we switch to this view to reset state.
        cls._target_col = None


    @classmethod
    def redraw(cls, en_changes:bool) -> None:
        # Called when redraw required.
        # en_changes: bool indicating if displayed entries need updating too
        if cls._year_viewed != View._cursor_date.year:
            cls._draw_year()
            cls._last_cursor = None
            en_changes = True
        cls._show_cursor()
        cls._show_datelabel()
        # Queue delayed redraw of day content
        if not cls._show_datecontent_pending:
            cls._date_content.foreach(Gtk.Widget.destroy)
            cls._show_datecontent_pending = True
            cls._date_content_count = 0
            # Schedule idle to add datecontent
            # Priority below draw, so datelabel will be redrawn while moving
            GLib.idle_add(cls._show_datecontent,priority=GLib.PRIORITY_HIGH_IDLE+40)
        if en_changes:
            GLib.idle_add(cls._show_gridcontent,priority=GLib.PRIORITY_HIGH_IDLE+35)


    @classmethod
    def _show_datelabel(cls) -> None:
        # Show cursor date in label in bottom panel of view.
        # Called on redraw
        st = View._cursor_date.strftime(GUI.date_formatting_text_noyear)+':'
        cls._date_label.set_text(st)


    @classmethod
    def _show_datecontent(cls) -> None:
        # Show events for current cursor date in bottom panel of view.
        # Assumes that no events are currently shown.
        # Can be slow, so called in idle from redraw.
        dt = View._cursor_date
        cls._visible_occurrences = Calendar.occurrence_list(dt, dt+timedelta(days=1))
        r = 0
        for occ in cls._visible_occurrences:
            if not cls.show_todos and isinstance(occ[0], iTodo):
                continue
            occ_dt_sta,occ_dt_end = start_end_dts_occ(occ)
            row = Gtk.Box()
            # Create entry mark (bullet or time) & add to row
            mark_label = cls.marker_label(occ[0], occ_dt_sta)
            ctx = mark_label.get_style_context()
            ctx.add_class('yearview_marker') # add style for CSS
            row.add(mark_label)
            # Create entry content label & add to row
            cont_label = cls.entry_text_label(occ[0], occ_dt_sta, occ_dt_end, add_location=cls._show_location, loc_max_chars=cls._loc_max_chars)
            cont_label.set_hexpand(True) # Also sets hexpand_set to True
            row.add(cont_label)
            cls._date_content.add(row)
            # See if we've hit the cursor target entry
            if cls._target_entry is not None and cls._target_entry is occ[0]:
                View._cursor_idx_in_date = r
                cls._target_entry = None
            r += 1
        cls._date_content_count = r
        cls._date_content.show_all()
        cls._show_entry_cursor()
        cls._show_datecontent_pending = False
        cls._target_entry = None # just in case - should be done already


    @classmethod
    def _show_entry_cursor(cls) -> None:
        # Set style to display entry cursor (= cursor in lower section of view)
        cch = GUI.create_events
        if cls._date_content_count==0:
            # No entries for day - return
            View._cursor_idx_in_date = 0
            GUI.set_menu_elts(on_event=False, can_create_here=cch) # Menu items
            return
        i = View._cursor_idx_in_date
        if i<0 or i>=cls._date_content_count:
            i = cls._date_content_count-1
            View._cursor_idx_in_date = i
        mk = cls._date_content.get_children()[i].get_children()[0]
        ctx = mk.get_style_context()
        ctx.add_class(cls.ENTRY_CURSOR_STYLE)
        cls._last_entry_cursor = i
        cls._scroll_to_cursor_required = True # to be read in draw handler
        # Enable/disable menu items
        on_ev = False
        on_td = False
        en = cls.get_cursor_entry()
        if isinstance(en, iTodo):
            on_td = True
        else:
            on_ev = True
        ro = Calendar.calendar_readonly(en)
        GUI.set_menu_elts(on_event=on_ev, on_todo=on_td, read_only=ro, can_create_here=cch)


    @classmethod
    def _pre_datecontent_draw(cls, wid:Gtk.Widget, _) -> bool:
        # Callback called on 'draw' event on date_content.
        # Called before drawing date content.
        # Used to scroll window when cursor has been moved (since we
        # need to have calculated the layout to know where to scoll to).
        if cls._scroll_to_cursor_required:
            cls.scroll_to_row(cls._date_content, View._cursor_idx_in_date, cls._date_content_scroll)
            cls._scroll_to_cursor_required = False
        return False # propagate event


    @classmethod
    def _hide_entry_cursor(cls) -> None:
        # Remove style from entry cursor (= cursor in lower section of view)
        if cls._last_entry_cursor is not None:
            mk = cls._date_content.get_children()[cls._last_entry_cursor].get_children()[0]
            ctx = mk.get_style_context()
            ctx.remove_class(cls.ENTRY_CURSOR_STYLE)
            cls._last_entry_cursor = None


    @classmethod
    def cursor_set_date(cls, dt:dt_date, idx:int=0, reset_target_col:bool=True) -> bool:
        # Set current cursor date & index in date.
        # Override default method so we can update state of target column.
        # Return True to indicate view supports jumping to a date.
        View._cursor_date = dt
        View._cursor_idx_in_date = idx
        if reset_target_col:
            cls._target_col = None
        return True


    @classmethod
    def get_cursor_entry(cls) -> iCal.Event:
        # Returns entry at cursor position, or None if cursor not on entry.
        # Called from cursor_edit_entry() & delete_request().
        if cls._date_content_count == 0:
            return None
        return cls._visible_occurrences[View._cursor_idx_in_date][0]


    @staticmethod
    def _local_date(dt:dt_date) -> dt_date:
        # Given a date/datetime, return the corresponding local date.
        # Note this is timezone aware, so 1am-tomorrow in another
        # timezone might correspond to today in local timezone.
        # Helper function for _show_gridcontent().
        if not isinstance(dt, dt_datetime):
            return dt
        if dt.tzinfo is None:
            return dt.date()
        return dt.astimezone().date()


    @classmethod
    def _show_gridcontent(cls) -> None:
        # Set styles to show events in grid.
        # Can be slow, so called in idle from redraw.
        # Potential optimisations here. Rather than getting occurences and
        # then splitting, we can add a custom calendar function to return
        # just the dates. Particularly for the repeated entries.
        yr = cls._year_viewed
        date = dt_date(year=yr,month=1,day=1)
        oneday = timedelta(days=1)
        single_list = Calendar.occurrence_list(date, dt_date(year=yr+1,month=1,day=1), include_single=True, include_repeated=False, in_grid=True)
        occ_dates_single = [cls._local_date(o[1]) for o in single_list if isinstance(o[0],iEvent)]
        if cls.show_todos:
            occ_dates_todo = [cls._local_date(o[1]) for o in single_list if isinstance(o[0],iTodo)]
        reps_list = Calendar.occurrence_list(date, dt_date(year=yr+1,month=1,day=1), include_single=False, include_repeated=True, in_grid=True)
        # Now simplify this list to keep the info we need...
        reps_list = [(o[0]['RRULE']['FREQ'][0],cls._local_date(o[1])) for o in reps_list if 'FREQ' in o[0]['RRULE']]
        occ_dates_repeated = [o[1] for o in reps_list]
        occ_dates_repeated_year = [o[1] for o in reps_list if o[0]=='YEARLY']
        occ_dates_repeated_month = [o[1] for o in reps_list if o[0]=='MONTHLY']
        occ_dates_repeated_week = [o[1] for o in reps_list if o[0]=='WEEKLY']
        occ_dates_repeated_day = [o[1] for o in reps_list if o[0]=='DAILY']
        occ_dates_repeated_hour = [o[1] for o in reps_list if o[0]=='HOURLY']
        occ_dates_repeated_minute = [o[1] for o in reps_list if o[0]=='MINUTELY']
        occ_dates_repeated_second = [o[1] for o in reps_list if o[0]=='SECONDLY']

        for m in range(1,13):
            day,daycount = calendar.monthrange(yr,m)
            x,y = cls._date_to_cell(date)
            for d in range(daycount):
                l = cls._grid_cells.get_child_at(x,y)
                ctx = l.get_style_context()
                if date in occ_dates_single:
                    ctx.add_class('yearview_entry_single')
                else:
                    ctx.remove_class('yearview_entry_single')
                if cls.show_todos:
                    if date in occ_dates_todo:
                        ctx.add_class('yearview_entry_todo')
                    else:
                        ctx.remove_class('yearview_entry_todo')
                ctx.remove_class('yearview_entry_repeated')
                ctx.remove_class('yearview_entry_repeated_year')
                ctx.remove_class('yearview_entry_repeated_month')
                ctx.remove_class('yearview_entry_repeated_week')
                ctx.remove_class('yearview_entry_repeated_day')
                ctx.remove_class('yearview_entry_repeated_hour')
                ctx.remove_class('yearview_entry_repeated_minute')
                ctx.remove_class('yearview_entry_repeated_second')
                if date in occ_dates_repeated:
                    ctx.add_class('yearview_entry_repeated')
                    if date in occ_dates_repeated_year:
                        ctx.add_class('yearview_entry_repeated_year')
                    if date in occ_dates_repeated_month:
                        ctx.add_class('yearview_entry_repeated_month')
                    if date in occ_dates_repeated_week:
                        ctx.add_class('yearview_entry_repeated_week')
                    if date in occ_dates_repeated_day:
                        ctx.add_class('yearview_entry_repeated_day')
                    if date in occ_dates_repeated_hour:
                        ctx.add_class('yearview_entry_repeated_hour')
                    if date in occ_dates_repeated_minute:
                        ctx.add_class('yearview_entry_repeated_minute')
                    if date in occ_dates_repeated_second:
                        ctx.add_class('yearview_entry_repeated_second')
                date += oneday
                x += 1


    @classmethod
    def _show_cursor(cls) -> None:
        # Add style class for grid cursor to the appropriate cell label
        new_coords = cls._date_to_cell(View._cursor_date)
        if cls._last_cursor!=new_coords:
            cls._hide_cursor()
        l = cls._grid_cells.get_child_at(*new_coords)
        ctx = l.get_style_context()
        ctx.add_class(cls.GRID_CURSOR_STYLE)
        cls._last_cursor = new_coords


    @classmethod
    def _hide_cursor(cls) -> None:
        # Remove grid cursor style class so grid cursor is not shown
        if cls._last_cursor is not None:
            l = cls._grid_cells.get_child_at(*cls._last_cursor)
            ctx = l.get_style_context()
            ctx.remove_class(cls.GRID_CURSOR_STYLE)
            cls._last_cursor = None


    @classmethod
    def _cursor_move_lr(cls, d:int) -> None:
        # Callback for user moving grid cursor left (d<0) or right (d>0).
        cls.cursor_inc(timedelta(days=d), 0)
        cls._target_col = None
        View._today_toggle_date = None


    @classmethod
    def _cursor_move_up(cls) -> None:
        # Callback for user moving grid cursor up.
        cur_dt = View._cursor_date
        if cls._target_col is None:
            cls._target_col = cls._date_to_cell(cur_dt)[0]
        new_y = cur_dt.month-2 # months 1..12, cells 0..11
        new_yr = cur_dt.year
        if new_y<0:
            new_y = 11
            new_yr -= 1
        new_dt = cls._cell_to_date_clamped(cls._target_col, new_y, new_yr)
        cls.cursor_set_date(new_dt, reset_target_col=False)
        cls.redraw(en_changes=False)
        View._today_toggle_date = None


    @classmethod
    def _cursor_move_dn(cls) -> None:
        # Callback for user moving grid cursor down.
        cur_dt = View._cursor_date
        if cls._target_col is None:
            cls._target_col = cls._date_to_cell(cur_dt)[0]
        new_y = cur_dt.month # months 1..12, cells 0..11
        new_yr = cur_dt.year
        if new_y>11:
            new_y = 0
            new_yr += 1
        new_dt = cls._cell_to_date_clamped(cls._target_col, new_y, new_yr)
        cls.cursor_set_date(new_dt, reset_target_col=False)
        cls.redraw(en_changes=False)
        View._today_toggle_date = None


    @classmethod
    def _cursor_move_stmon(cls) -> None:
        # Callback for user moving grid cursor to start of month.
        new_dt = View._cursor_date.replace(day=1)
        cls.cursor_set_date(new_dt, reset_target_col=True)
        cls.redraw(en_changes=False)
        View._today_toggle_date = None


    @classmethod
    def _cursor_move_endmon(cls) -> None:
        # Callback for user moving grid cursor to end of month.
        dt = View._cursor_date
        lastday = calendar.monthrange(dt.year,dt.month)[1]
        new_dt = dt.replace(day=lastday)
        cls.cursor_set_date(new_dt, reset_target_col=True)
        cls.redraw(en_changes=False)
        View._today_toggle_date = None


    @classmethod
    def _cursor_move_pgupdn(cls, d:int) -> None:
        # Callback for user moving grid cursor pageup (d<0) or pagedown (d>0).
        cur_dt = View._cursor_date
        if cls._target_col is None:
            cls._target_col = cls._date_to_cell(cur_dt)[0]
        new_yr = cur_dt.year+d
        new_dt = cls._cell_to_date_clamped(cls._target_col, cur_dt.month-1, new_yr)
        cls.cursor_set_date(new_dt, reset_target_col=False)
        cls.redraw(en_changes=False)
        View._today_toggle_date = None


    @classmethod
    def _cursor_move_today(cls) -> None:
        # Callback for user toggling cursors between today & another date
        today = dt_date.today()
        if View._cursor_date != today:
            View._today_toggle_date = View._cursor_date
            View._today_toggle_idx = View._cursor_idx_in_date
            cls.cursor_set_date(today, reset_target_col=True)
            cls.redraw(en_changes=False)
        elif View._today_toggle_date is not None:
            cls.cursor_set_date(View._today_toggle_date, idx=View._today_toggle_idx, reset_target_col=True)
            cls.redraw(en_changes=False)


    @classmethod
    def _idxcursor_move_up(cls) -> None:
        # Callback to move entry/index cursor (=cursor in bottom section) up
        if View._cursor_idx_in_date>0:
            View._cursor_idx_in_date -= 1
            cls._hide_entry_cursor()
            cls._show_entry_cursor()


    @classmethod
    def _idxcursor_move_dn(cls) -> None:
        # Callback to move entry/index cursor (=cursor in bottom section) down
        if View._cursor_idx_in_date<cls._date_content_count-1:
            View._cursor_idx_in_date += 1
            cls._hide_entry_cursor()
            cls._show_entry_cursor()


    @classmethod
    def keypress(cls, wid:Gtk.Widget, ev:Gdk.EventKey) -> None:
        # Handle key press event in Year view.
        # Called (from GUI.keypress()) on keypress (or repeat) event
        if ev.state & Gdk.ModifierType.SHIFT_MASK:
            # Shift key is being pressed
            try:
                f = cls._KEYMAP_SHIFT[ev.keyval]
                GLib.idle_add(f, priority=GLib.PRIORITY_HIGH_IDLE+30)
                return # If it gets here without errors, we're done
            except KeyError:
                pass
        try:
            f = cls._KEYMAP[ev.keyval]
            GLib.idle_add(f, priority=GLib.PRIORITY_HIGH_IDLE+30)
            return # If it gets here without errors, we're done
        except KeyError:
            pass
        if GUI.create_events and ev.state & (Gdk.ModifierType.CONTROL_MASK|Gdk.ModifierType.MOD1_MASK)==0 and Gdk.KEY_exclam <= ev.keyval <= Gdk.KEY_asciitilde:
            date = cls.cursor_date()
            GLib.idle_add(EventDialogController.new_event, chr(ev.keyval), date, priority=GLib.PRIORITY_HIGH_IDLE+30)


    @classmethod
    def click_grid(cls, wid:Gtk.Widget, ev:Gdk.EventButton) -> bool:
        # Callback. Called whenever day grid is clicked/tapped.
        # Move grid (main) cursor to cell that was clicked.
        base_cwid,ex_count = divmod(wid.get_allocated_width(), cls.GRID_COLUMNS)
        # GTK will layout the left-most ex_count cells with width base_cwid+1.
        # The remaining cells with width base_cwid.
        wid_ex = (base_cwid+1)*ex_count # total width of left-side "wide" cells
        if ev.x <= wid_ex:
            x = int(ev.x/(base_cwid+1))
        else:
            x = ex_count + int((ev.x-wid_ex)/base_cwid)
        y = int(cls.GRID_ROWS*ev.y/wid.get_allocated_height())
        dt = cls._cell_to_date_clamped(x, y, cls._year_viewed)
        GLib.idle_add(cls._jump_to_date, dt, priority=GLib.PRIORITY_HIGH_IDLE+30)
        return True # event handled - don't propagate


    @classmethod
    def click_events(cls, wid:Gtk.Widget, ev:Gdk.EventButton) -> bool:
        # Callback. Called whenever events area at bottom is clicked/tapped.
        # Move entry cursor to entry that was clicked
        new_idx = cls.y_to_day_row(cls._date_content, ev.y, cls._date_content_count, cls._date_content_scroll)
        if View._cursor_idx_in_date != new_idx:
            View._cursor_idx_in_date = new_idx
            cls._hide_entry_cursor()
            cls._show_entry_cursor()
        return True # event handled - don't propagate


    @classmethod
    def _jump_to_date(cls, dt:dt_date) -> None:
        # Idle callback to move grid cursor to given date, e.g. on click
        cls.cursor_set_date(dt, reset_target_col=True)
        cls.redraw(en_changes=False)


    @staticmethod
    def _date_to_cell(dt:dt_date) -> Tuple[int,int]:
        # Helper function to return cell coordinates of given date.
        stday = calendar.monthrange(dt.year,dt.month)[0] # start day of month
        x = dt.day-1+(stday-Config.get_int('global','start_week_day'))%7
        y = dt.month-1
        return x,y


    @staticmethod
    def _cell_to_date_clamped(x:int, y:int, yr:int) -> dt_date:
        # Helper function to return date given cell coordinates.
        # "Clamps" date, so if the cell is before the start or after the end
        # of the month then it returns the closest date in the month.
        m = y+1
        st_wk = Config.get_int('global','start_week_day')
        st_day,monthdays = calendar.monthrange(yr,m)
        d = x+1-(st_day-st_wk)%7
        if d<1:
            d = 1
        else:
            d = min(d,monthdays)
        return dt_date(year=yr, month=m, day=d)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# test_import_paste.py
# Unit tests for importing/pasting events.
# (Both of these use the Calendar.new_entry_from_example() function.)
#
# Copyright (C) 2024 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.
#

import unittest
from datetime import date, time, datetime, timezone, timedelta
from dateutil import tz
from os import remove as os_remove
from icalendar import Calendar as iCalendar, Event as iEvent, Todo as iTodo, Alarm as iAlarm

# Add '..' to path, so this can be run from test directory
import sys
sys.path.append('..')

# Import Pygenda modules...
from pygenda.pygenda_calendar import Calendar
from pygenda.pygenda_config import Config
from pygenda.pygenda_entryinfo import EntryInfo
from pygenda.pygenda_util import get_local_tz, _set_local_tz as set_local_tz


class TestImportPaste(unittest.TestCase):
    maxDiff = None # show unlimited chars when showing diffs
    TESTFILE_NAME = 'test_import_paste_TESTFILE.ics'
    UID_BASE = 'Pygenda-test-import-paste-5415-'
    uid_index = 0

    @classmethod
    def setUpClass(cls):
        # Called once before all tests
        # Override config options so it uses our test ics file
        Config.set('calendar', 'type', 'icalfile')
        Config.set('calendar', 'filename', cls.TESTFILE_NAME)
        Config.set('calendar', 'display_name', 'Test calendar for test_entries')
        Config.set('calendar', 'readonly', None)
        Config.set('calendar', 'entry_type', None)
        Config.set('calendar1', 'type', None) # so only specified file opened

        # Save local timezone, so running tests will test in your tz
        cls.local_tz_saved = get_local_tz()


    def setUp(self) -> None:
        # This is called before each individual test function

        # Reset timezone for next test (might be changed in test)
        set_local_tz(self.local_tz_saved)

        # Delete test ics files, so an empty one is created
        self._delete_testfile()

        # Call init to create the calendar
        Calendar.init()


    @classmethod
    def tearDownClass(cls) -> None:
        # This is called after final test
        cls._delete_testfile()


    @classmethod
    def _delete_testfile(cls) -> None:
        # Helper function for setup/teardown
        try:
            os_remove(cls.TESTFILE_NAME)
        except FileNotFoundError:
            pass


    @staticmethod
    def _past_timestamp() -> datetime:
        # Return a time in the past for use as a timestamp
        t =  datetime.now(timezone.utc).replace(microsecond=0)
        t -= timedelta(seconds=2)
        return t


    @classmethod
    def _new_event(cls, desc:str, start:date, end:date=None, status:str=None, rfreq:str=None, longdesc:str=None, location:str=None, alarm:iAlarm=None) -> iEvent:
        # Helper function to create an event
        ev = iEvent()
        ev.add('SUMMARY', desc)
        # we need to read & store tzid from start date for these tests.
        # (This should be done in Pygenda when it creates/imports events.)
        st_tznm = cls._get_tzname_from_datetime(start)
        ev.add('DTSTART', start, parameters={'TZID':st_tznm} if st_tznm else None)
        if end is not None:
            end_tznm = cls._get_tzname_from_datetime(end)
            ev.add('DTEND', end, parameters={'TZID':end_tznm} if end_tznm else None)
        if status is not None:
            ev.add('STATUS', status)
        if rfreq is not None:
            ev.add('RRULE', {'FREQ':[rfreq]})
        if longdesc is not None:
            ev.add('DESCRIPTION', longdesc)
        if location is not None:
            ev.add('LOCATION', location)
        if alarm is not None:
            ev.add_component(alarm)

        ev.add('UID', cls.UID_BASE+str(cls.uid_index))
        cls.uid_index += 1
        stamp = cls._past_timestamp()
        ev.add('CREATED', stamp)
        ev.add('DTSTAMP', stamp)
        return ev


    @classmethod
    def _new_todo(cls, desc:str, cats:tuple=(), due:date=None, priority:int=None, status:str=None, longdesc:str=None) -> iTodo:
        # Helper function to create a todo
        if type(cats) != tuple:
            raise(ValueError('Error - cats needs to be a tuple'))
        td = iTodo()
        td.add('SUMMARY', desc)
        td.add('UID', cls.UID_BASE+str(cls.uid_index))
        cls.uid_index += 1
        stamp = cls._past_timestamp()
        td.add('CREATED', stamp)
        td.add('DTSTAMP', stamp)
        if cats:
            td.add('CATEGORIES', list(cats))
        if due is not None:
            td.add('DUE', due)
        if priority is not None:
            td.add('PRIORITY', priority)
        if status is not None:
            td.add('STATUS', status)
        if longdesc is not None:
            td.add('DESCRIPTION', longdesc)
        return td


    @staticmethod
    def _get_tzname_from_datetime(dt) -> str:
        # Helper fn return a string with the name of the timezone.
        # Returns None if no timezone, unrecognized tz, or dt is a date.
        tzname = None
        if isinstance(dt, datetime) and dt.tzinfo is not None:
            tzinfo_str = str(dt.tzinfo)
            idx = -1
            for i in range(2): # Find second-to-last '/'
                idx = tzinfo_str.rfind('/', 0, idx)
                if idx == -1: # '/' not found
                    break
            if idx != -1:
                tzname = tzinfo_str[idx+1:-2]
        return tzname


    @staticmethod
    def _do_import(en):
        # Perform import call as would happen in application
        Calendar.import_entry(en, cal_idx=0)


    @staticmethod
    def _do_paste_event(en, targ_date:date):
        # Perform paste call as would happen in "event" views
        Calendar.paste_entry(en, e_type=EntryInfo.TYPE_EVENT, dt_start=targ_date)


    @staticmethod
    def _do_paste_todo(en, targ_cats:list=None):
        # Perform paste call as would happen in Todo view
        Calendar.paste_entry(en, e_type=EntryInfo.TYPE_TODO, e_cats=targ_cats)


    def _get_saved_version(self, en):
        # Read and return version of entry from file
        with open(self.TESTFILE_NAME, 'rb') as file:
            cal = iCalendar.from_ical(file.read())

        for ev in cal.walk('VEVENT'):
            if ev['UID'] == en['UID']:
                return ev

        for td in cal.walk('VTODO'):
            if td['UID'] == en['UID']:
                return td

        return None


    #@unittest.skip
    def test_import_01_event_simple(self) -> None:
        # Import simple event
        SHORT_DESC = 'Test event'
        DT_ST = date(2020,1,5)
        ev = self._new_event(SHORT_DESC, DT_ST)
        self._do_import(ev)

        # Test imported entry
        l = Calendar.occurrence_list(date(2020,1,1), date(2021,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_ST)
        imported = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['DTSTART'].dt, DT_ST)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(imported)
        self.check_stamps(imported, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_ST)


    #@unittest.skip
    def test_import_02_event_timed(self) -> None:
        # Import basic timed event
        SHORT_DESC = 'Test event timed'
        DT_ST = datetime(2021,2,5,12,13,14)
        ev = self._new_event(SHORT_DESC, DT_ST)
        self._do_import(ev)

        # Test imported entry
        l = Calendar.occurrence_list(date(2021,1,1), date(2022,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_ST)
        imported = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['DTSTART'].dt, DT_ST)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(imported)
        self.check_stamps(imported, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_ST)


    #@unittest.skip
    def test_import_03_event_time_end(self) -> None:
        # Import event with start & end times
        SHORT_DESC = 'Test event start & end times'
        DT_ST = datetime(2021,4,21,11,18)
        DT_END = datetime(2021,4,21,14,32)
        ev = self._new_event(SHORT_DESC, DT_ST, end=DT_END)
        self._do_import(ev)

        # Test imported entry
        l = Calendar.occurrence_list(date(2021,1,1), date(2022,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_ST)
        imported = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['DTSTART'].dt, DT_ST)
        self.assertEqual(imported['DTEND'].dt, DT_END)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(imported)
        self.check_stamps(imported, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_ST)
        self.assertEqual(ev_saved['DTEND'].dt, DT_END)


    #@unittest.skip
    def test_import_04_event_day(self) -> None:
        # Import day event
        SHORT_DESC = 'Test day event'
        DT_ST = date(1981,5,1)
        DT_END = date(1981,5,2)
        ev = self._new_event(SHORT_DESC, DT_ST, end=DT_END)
        self._do_import(ev)

        # Test imported entry
        l = Calendar.occurrence_list(date(1981,1,1), date(1982,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_ST)
        imported = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['DTSTART'].dt, DT_ST)
        self.assertEqual(imported['DTEND'].dt, DT_END)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(imported)
        self.check_stamps(imported, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_ST)
        self.assertEqual(ev_saved['DTEND'].dt, DT_END)


    #@unittest.skip
    def test_import_05_event_timezone(self) -> None:
        # Import timezoned event
        SHORT_DESC = 'Test event timezone NY'
        tz_NY = tz.gettz('America/New_York')
        self.assertTrue(tz_NY) # check not None
        DT_ST = datetime(2023,7,3,0,30,tzinfo=tz_NY)
        ev = self._new_event(SHORT_DESC, DT_ST)
        self._do_import(ev)

        # Test imported entry
        l = Calendar.occurrence_list(date(2023,1,1), date(2024,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_ST)
        imported = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['DTSTART'].dt, DT_ST)
        self.assertEqual(imported['DTSTART'].params['TZID'], 'America/New_York')

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(imported)
        self.check_stamps(imported, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_ST)
        self.assertEqual(ev_saved['DTSTART'].params['TZID'], 'America/New_York')


    #@unittest.skip
    def test_import_06_event_timezone_end(self) -> None:
        # Import timezoned event with timezoned end time
        SHORT_DESC = 'Test event timezone Brazil'
        tz_AC = tz.gettz('Brazil/Acre')
        self.assertTrue(tz_AC) # check not None
        DT_ST = datetime(2024,2,3,10,30,tzinfo=tz_AC)
        DT_END = datetime(2024,2,3,11,30,tzinfo=tz_AC)
        ev = self._new_event(SHORT_DESC, DT_ST, end=DT_END)
        self._do_import(ev)

        # Test imported entry
        l = Calendar.occurrence_list(date(2024,1,1), date(2025,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_ST)
        imported = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['DTSTART'].dt, DT_ST)
        self.assertEqual(imported['DTEND'].dt, DT_END)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(imported)
        self.check_stamps(imported, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_ST)
        self.assertEqual(ev_saved['DTSTART'].params['TZID'], 'Brazil/Acre')
        self.assertEqual(ev_saved['DTEND'].dt, DT_END)
        self.assertEqual(ev_saved['DTEND'].params['TZID'], 'Brazil/Acre')


    #@unittest.skip
    def test_import_07_event_status(self) -> None:
        # Import event with a status
        SHORT_DESC = u'Test event with status'
        DT_ST = datetime(2014,10,1,23,59)
        STATUS = 'CANCELLED'
        ev = self._new_event(SHORT_DESC, DT_ST, status=STATUS)
        self._do_import(ev)

        # Test imported entry
        l = Calendar.occurrence_list(date(2014,1,1), date(2015,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_ST)
        imported = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['DTSTART'].dt, DT_ST)
        self.assertEqual(imported['STATUS'], STATUS)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(imported)
        self.check_stamps(imported, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_ST)
        self.assertEqual(ev_saved['STATUS'], STATUS)


    #@unittest.skip
    def test_import_08_event_longdesc(self) -> None:
        # Import event with a long description
        SHORT_DESC = u'Test event with long desc & non-ascii like θ'
        LONG_DESC = u"""Big long description with some newlines
Like that one
and non-ascii characters like ë & ☉
"""
        DT_ST = date(2023,9,9)
        ev = self._new_event(SHORT_DESC, DT_ST, longdesc=LONG_DESC)
        self._do_import(ev)

        # Test imported entry
        l = Calendar.occurrence_list(date(2023,1,1), date(2024,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_ST)
        imported = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['DTSTART'].dt, DT_ST)
        self.assertEqual(imported['DESCRIPTION'], LONG_DESC)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(imported)
        self.check_stamps(imported, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_ST)
        self.assertEqual(ev_saved['DESCRIPTION'], LONG_DESC)


    #@unittest.skip
    def test_import_09_event_location(self) -> None:
        # Import event with a location
        SHORT_DESC = 'Test event with a location'
        LOCATION = 'Liverpool'
        DT_ST = date(2026,1,8)
        ev = self._new_event(SHORT_DESC, DT_ST, location=LOCATION)
        self._do_import(ev)

        # Test imported entry
        l = Calendar.occurrence_list(date(2026,1,1), date(2027,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_ST)
        imported = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['DTSTART'].dt, DT_ST)
        self.assertEqual(imported['LOCATION'], LOCATION)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(imported)
        self.check_stamps(imported, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_ST)
        self.assertEqual(ev_saved['LOCATION'], LOCATION)


    #@unittest.skip
    def test_import_10_event_repeating(self) -> None:
        # Import repeating event
        SHORT_DESC = 'Test repeating event'
        DT_ST = date(1989,6,19)
        ev = self._new_event(SHORT_DESC, DT_ST, rfreq="YEARLY")
        self._do_import(ev)

        # Test imported entry
        l = Calendar.occurrence_list(date(1980,1,1), date(1989,1,1))
        self.assertEqual(len(l), 0)
        l = Calendar.occurrence_list(date(1989,1,1), date(2010,1,1))
        self.assertEqual(len(l), 21)
        l = Calendar.occurrence_list(date(1989,1,1), date(1990,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_ST)
        imported = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['DTSTART'].dt, DT_ST)
        self.assertEqual(imported['RRULE']['FREQ'][0], 'YEARLY')

        l = Calendar.occurrence_list(date(1991,1,1), date(1992,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_ST.replace(year=1991))

        l = Calendar.occurrence_list(date(2005,1,1), date(2006,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_ST.replace(year=2005))

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(imported)
        self.check_stamps(imported, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_ST)
        self.assertEqual(ev_saved['RRULE']['FREQ'][0], 'YEARLY')


    #@unittest.skip
    def test_import_11_event_alarm(self) -> None:
        # Import event with alarm
        # Note: current behaviour is to discard the alarm!!
        SHORT_DESC = 'Test event with alarm'
        DT_ST = datetime(2025,2,14,10,20)
        alm = iAlarm()
        alm.add('TRIGGER', timedelta(minutes=-10))
        alm.add('ACTION', 'AUDIO')
        ev = self._new_event(SHORT_DESC, DT_ST, alarm=alm)
        self.assertEqual(len(ev.walk('VALARM')), 1)
        self._do_import(ev)

        # Test imported entry
        l = Calendar.occurrence_list(date(2025,1,1), date(2026,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_ST)
        imported = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['DTSTART'].dt, DT_ST)
        self.assertEqual(len(imported.walk('VALARM')), 0)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(imported)
        self.check_stamps(imported, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_ST)
        self.assertEqual(len(ev_saved.walk('VALARM')), 0)


    #@unittest.skip
    def test_import_12_todo_simple(self) -> None:
        # Import simple todo
        SHORT_DESC = 'Test todo'
        td = self._new_todo(SHORT_DESC)
        self._do_import(td)

        # Test imported entry
        l = Calendar.todo_list()
        self.assertEqual(len(l), 1)
        imported = l[0] # Simple list (not occs) for todos
        self.check_stamps(td, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)

        # Test version saved to disk is also OK
        td_saved = self._get_saved_version(imported)
        self.check_stamps(imported, td_saved, same_obj=True)
        self.assertEqual(td_saved['SUMMARY'], SHORT_DESC)


    #@unittest.skip
    def test_import_13_todo_wipe_categories(self) -> None:
        # Test todo import will remove categories
        SHORT_DESC = 'Test import todo with categories'
        td = self._new_todo(SHORT_DESC, cats=('cat1',))
        self.assertEqual(td['CATEGORIES'].cats[0], 'cat1')
        self._do_import(td)

        # Test imported entry
        l = Calendar.todo_list()
        self.assertEqual(len(l), 1)
        imported = l[0]
        self.check_stamps(td, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertNotIn('CATEGORIES', imported)

        # Test version saved to disk is also OK
        td_saved = self._get_saved_version(imported)
        self.check_stamps(imported, td_saved, same_obj=True)
        self.assertEqual(td_saved['SUMMARY'], SHORT_DESC)
        self.assertNotIn('CATEGORIES', td_saved)


    #@unittest.skip
    def test_import_14_todo_priority_status(self) -> None:
        # Import todo with priority & status
        SHORT_DESC = 'Test todo with priority & status'
        PRIORITY = 3
        STATUS = 'IN-PROCESS'
        td = self._new_todo(SHORT_DESC, priority=PRIORITY, status=STATUS)
        self._do_import(td)

        # Test imported entry
        l = Calendar.todo_list()
        self.assertEqual(len(l), 1)
        imported = l[0] # Simple list (not occs) for todos
        self.check_stamps(td, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['PRIORITY'], PRIORITY)
        self.assertEqual(imported['STATUS'], STATUS)

        # Test version saved to disk is also OK
        td_saved = self._get_saved_version(imported)
        self.check_stamps(imported, td_saved, same_obj=True)
        self.assertEqual(td_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(td_saved['PRIORITY'], PRIORITY)
        self.assertEqual(td_saved['STATUS'], STATUS)


    #@unittest.skip
    def test_import_15_todo_duedate(self) -> None:
        # Import todo with due date
        SHORT_DESC = 'Test todo with due date'
        DT_DUE = date(2018,12,24)
        td = self._new_todo(SHORT_DESC, due=DT_DUE)
        self._do_import(td)

        # Test imported entry
        l = Calendar.todo_list()
        self.assertEqual(len(l), 1)
        imported = l[0] # Simple list (not occs) for todos
        self.check_stamps(td, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['DUE'].dt, DT_DUE)

        # Test version saved to disk is also OK
        td_saved = self._get_saved_version(imported)
        self.check_stamps(imported, td_saved, same_obj=True)
        self.assertEqual(td_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(td_saved['DUE'].dt, DT_DUE)


    #@unittest.skip
    def test_import_16_todo_longdesc(self) -> None:
        # Import todo with longdesc
        SHORT_DESC = 'Test todo with long description'
        LONG_DESC = u'Lørem ipsum dolor sit åmet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ût enim ad minim veñam, quis nostrud exercitation üllamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit eße cillum dolore eu fugiat nulla pariatur. Excepteur sint occæcat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.'
        td = self._new_todo(SHORT_DESC, longdesc=LONG_DESC)
        self._do_import(td)

        # Test imported entry
        l = Calendar.todo_list()
        self.assertEqual(len(l), 1)
        imported = l[0] # Simple list (not occs) for todos
        self.check_stamps(td, imported)
        self.assertEqual(imported['SUMMARY'], SHORT_DESC)
        self.assertEqual(imported['DESCRIPTION'], LONG_DESC)

        # Test version saved to disk is also OK
        td_saved = self._get_saved_version(imported)
        self.check_stamps(imported, td_saved, same_obj=True)
        self.assertEqual(td_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(td_saved['DESCRIPTION'], LONG_DESC)


    #@unittest.skip
    def test_paste_01_event_simple(self) -> None:
        # Paste a simple event
        SHORT_DESC = 'Test event'
        DT_ST = date(2020,1,5)
        ev = self._new_event(SHORT_DESC, DT_ST)
        DT_TARG = date(2023,7,11)
        self._do_paste_event(ev, DT_TARG)

        # Test no entry exists where event was pasted from
        l = Calendar.occurrence_list(date(2020,1,1), date(2021,1,1))
        self.assertEqual(len(l), 0)

        # Test pasted entry
        l = Calendar.occurrence_list(date(2023,1,1), date(2024,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_TARG)
        pasted = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DTSTART'].dt, DT_TARG)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_TARG)


    #@unittest.skip
    def test_paste_02_event_time(self) -> None:
        # Paste event with a time
        SHORT_DESC = 'Test event with time'
        DT_ST = datetime(1993,9,30,12,35)
        ev = self._new_event(SHORT_DESC, DT_ST)
        DT_TARG = date(2024,1,1)
        self._do_paste_event(ev, DT_TARG)
        DT_WANT = datetime(2024,1,1,12,35)

        # Test no entry exists where event was pasted from
        l = Calendar.occurrence_list(date(1993,1,1), date(1994,1,1))
        self.assertEqual(len(l), 0)

        # Test pasted entry
        l = Calendar.occurrence_list(date(2024,1,1), date(2025,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_WANT)
        pasted = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DTSTART'].dt, DT_WANT)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_WANT)


    #@unittest.skip
    def test_paste_03_event_endtime(self) -> None:
        # Paste event with start & end time
        SHORT_DESC = 'Test event with start/end times'
        DT_ST = datetime(2010,2,28,23,15)
        DT_END = datetime(2010,3,1,0,30)
        ev = self._new_event(SHORT_DESC, DT_ST, end=DT_END)
        DT_TARG = date(1999,12,31)
        self._do_paste_event(ev, DT_TARG)
        DT_WANT_ST = datetime(1999,12,31,23,15)
        DT_WANT_END = datetime(2000,1,1,0,30)

        # Test no entry exists where event was pasted from
        l = Calendar.occurrence_list(date(2010,1,1), date(2011,1,1))
        self.assertEqual(len(l), 0)

        # Test pasted entry
        l = Calendar.occurrence_list(date(1999,1,1), date(2000,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_WANT_ST)
        pasted = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DTSTART'].dt, DT_WANT_ST)
        self.assertEqual(pasted['DTEND'].dt, DT_WANT_END)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_WANT_ST)
        self.assertEqual(ev_saved['DTEND'].dt, DT_WANT_END)


    #@unittest.skip
    def test_paste_04_event_multiday(self) -> None:
        # Paste multi-day event
        SHORT_DESC = 'Test 4-day event'
        DT_ST = date(2014,10,10)
        DT_END = date(2014,10,15)
        ev = self._new_event(SHORT_DESC, DT_ST, end=DT_END)
        DT_TARG = date(2015,1,31)
        self._do_paste_event(ev, DT_TARG)
        DT_WANT_END = date(2015,2,5)

        # Test no entry exists where event was pasted from
        l = Calendar.occurrence_list(date(2014,1,1), date(2015,1,1))
        self.assertEqual(len(l), 0)

        # Test pasted entry
        l = Calendar.occurrence_list(date(2015,1,1), date(2015,2,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_TARG)
        pasted = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DTSTART'].dt, DT_TARG)
        self.assertEqual(pasted['DTEND'].dt, DT_WANT_END)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_TARG)
        self.assertEqual(ev_saved['DTEND'].dt, DT_WANT_END)


    #@unittest.skip
    def test_paste_05_event_timezone(self) -> None:
        # Paste timezoned event
        SHORT_DESC = 'Test event timezone NY'
        tz_NY = tz.gettz('America/New_York')
        self.assertTrue(tz_NY) # check not None
        DT_ST = datetime(2023,7,3,18,30,tzinfo=tz_NY)
        ev = self._new_event(SHORT_DESC, DT_ST)
        DT_TARG = date(2023,7,5)
        self._do_paste_event(ev, DT_TARG)

        # Test pasted entry
        l = Calendar.occurrence_list(DT_TARG, DT_TARG+timedelta(days=1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1].time(), DT_ST.time())
        pasted = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DTSTART'].dt.time(), DT_ST.time())
        self.assertEqual(pasted['DTSTART'].params['TZID'], 'America/New_York')

        # Paste another - to winter
        DT_TARG2 = date(2023,12,13)
        self._do_paste_event(ev, DT_TARG2)

        # Should be two events in the year now
        l = Calendar.occurrence_list(date(2023,1,1), date(2024,1,1))
        self.assertEqual(len(l), 2)

        # Test second pasted entry
        l = Calendar.occurrence_list(DT_TARG2, DT_TARG2+timedelta(days=1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1].time(), DT_ST.time())
        pasted2 = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, pasted2, new_copy=True)
        self.assertNotEqual(pasted['UID'], pasted2['UID'])
        self.assertEqual(pasted2['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted2['DTSTART'].dt.time(), DT_ST.time())
        self.assertEqual(pasted2['DTSTART'].params['TZID'], 'America/New_York')

        # Test both versions saved to disk are OK
        ev_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, pasted['DTSTART'].dt)
        self.assertEqual(ev_saved['DTSTART'].params['TZID'], 'America/New_York')
        ev_saved2 = self._get_saved_version(pasted2)
        self.check_stamps(pasted2, ev_saved2, same_obj=True)
        self.assertEqual(ev_saved2['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved2['DTSTART'].dt, pasted2['DTSTART'].dt)
        self.assertEqual(ev_saved2['DTSTART'].params['TZID'],'America/New_York')

        # Test local times of all pasted entries
        set_local_tz(tz.gettz('Europe/Paris'))
        self.assertEqual(pasted['DTSTART'].dt.time(), time(18,30))
        self.assertEqual(pasted['DTSTART'].dt.astimezone(get_local_tz()).time(), time(0,30))
        self.assertEqual(pasted2['DTSTART'].dt.time(), time(18,30))
        self.assertEqual(pasted2['DTSTART'].dt.astimezone(get_local_tz()).time(), time(0,30))

        self.assertEqual(ev_saved['DTSTART'].dt.time(), time(18,30))
        self.assertEqual(ev_saved['DTSTART'].dt.astimezone(get_local_tz()).time(), time(0,30))
        self.assertEqual(ev_saved2['DTSTART'].dt.time(), time(18,30))
        self.assertEqual(ev_saved2['DTSTART'].dt.astimezone(get_local_tz()).time(), time(0,30))


    #@unittest.skip
    def test_paste_06_event_timezone_endtime(self) -> None:
        # Paste timezoned event
        SHORT_DESC = 'Test event timezone NY with endtime'
        tz_NY = tz.gettz('America/New_York')
        self.assertTrue(tz_NY) # check not None
        DT_ST = datetime(2000,3,3,0,59,tzinfo=tz_NY)
        DT_END = datetime(2000,5,17,23,1,tzinfo=tz_NY)
        ev = self._new_event(SHORT_DESC, DT_ST, end=DT_END)
        DT_TARG = date(2000,5,16)
        self._do_paste_event(ev, DT_TARG)

        # Test pasted entry
        l = Calendar.occurrence_list(DT_TARG, DT_TARG+timedelta(days=1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1].time(), DT_ST.time())
        pasted = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DTSTART'].dt.time(), DT_ST.time())
        self.assertEqual(pasted['DTSTART'].params['TZID'], 'America/New_York')
        self.assertEqual(pasted['DTEND'].dt.time(), DT_END.time())
        self.assertEqual(pasted['DTEND'].params['TZID'], 'America/New_York')

        # Test version saved to disk is OK
        ev_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, pasted['DTSTART'].dt)
        self.assertEqual(ev_saved['DTSTART'].params['TZID'], 'America/New_York')
        self.assertEqual(ev_saved['DTEND'].dt, pasted['DTEND'].dt)
        self.assertEqual(ev_saved['DTEND'].params['TZID'], 'America/New_York')

        # Test local times of all pasted entries
        set_local_tz(tz.gettz('Europe/Paris'))
        self.assertEqual(pasted['DTSTART'].dt.time(), time(0,59))
        self.assertEqual(pasted['DTSTART'].dt.astimezone(get_local_tz()).time(), time(6,59))
        self.assertEqual(ev_saved['DTSTART'].dt.time(), time(0,59))
        self.assertEqual(ev_saved['DTSTART'].dt.astimezone(get_local_tz()).time(), time(6,59))
        self.assertEqual(pasted['DTEND'].dt.time(), time(23,1))
        self.assertEqual(pasted['DTEND'].dt.astimezone(get_local_tz()).time(), time(5,1))
        self.assertEqual(ev_saved['DTEND'].dt.time(), time(23,1))
        self.assertEqual(ev_saved['DTEND'].dt.astimezone(get_local_tz()).time(), time(5,1))


    #@unittest.skip
    def test_paste_07_event_status(self) -> None:
        # Paste event with a status
        SHORT_DESC = u'Test event with status'
        DT_ST = date(2025,10,31)
        STATUS = 'TENTATIVE'
        ev = self._new_event(SHORT_DESC, DT_ST, status=STATUS)
        DT_TARG = date(2025,11,1)
        self._do_paste_event(ev, DT_TARG)

        # Test no entry exists where event was pasted from
        l = Calendar.occurrence_list(date(2025,10,1), date(2025,11,1))
        self.assertEqual(len(l), 0)

        # Test pasted entry
        l = Calendar.occurrence_list(date(2025,11,1), date(2025,12,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_TARG)
        pasted = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DTSTART'].dt, DT_TARG)
        self.assertEqual(pasted['STATUS'], STATUS)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_TARG)
        self.assertEqual(ev_saved['STATUS'], STATUS)


    #@unittest.skip
    def test_paste_08_event_longdesc(self) -> None:
        # Paste event with a long description
        SHORT_DESC = u'Test event with long desc & non-ascii like €'
        LONG_DESC = u"""Big long description with some newlines
Like this one
and non-ascii characters like é & â
"""
        DT_ST = datetime(2032,10,10,12,0)
        ev = self._new_event(SHORT_DESC, DT_ST, longdesc=LONG_DESC)
        DT_TARG = date(2124,2,29) # leap day
        DT_WANT_ST = datetime(2124,2,29,12,0)
        self._do_paste_event(ev, DT_TARG)

        # Test no entry exists where event was pasted from
        l = Calendar.occurrence_list(date(2032,1,1), date(2033,1,1))
        self.assertEqual(len(l), 0)

        # Test pasted entry
        l = Calendar.occurrence_list(date(2124,1,1), date(2125,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_WANT_ST)
        pasted = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DTSTART'].dt, DT_WANT_ST)
        self.assertEqual(pasted['DESCRIPTION'], LONG_DESC)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_WANT_ST)
        self.assertEqual(ev_saved['DESCRIPTION'], LONG_DESC)


    #@unittest.skip
    def test_paste_09_event_location(self) -> None:
        # Paste event with a location
        SHORT_DESC = u'Test event with a location'
        LOCATION = 'Liverpool'
        DT_ST = date(2026,1,8)
        ev = self._new_event(SHORT_DESC, DT_ST, location=LOCATION)
        DT_TARG = date(2026,1,10)
        self._do_paste_event(ev, DT_TARG)

        # Test no entry exists where event was pasted from
        l = Calendar.occurrence_list(date(2026,1,1), date(2026,1,10))
        self.assertEqual(len(l), 0)

        # Test pasted entry
        l = Calendar.occurrence_list(date(2026,1,1), date(2027,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_TARG)
        pasted = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DTSTART'].dt, DT_TARG)
        self.assertEqual(pasted['LOCATION'], LOCATION)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_TARG)
        self.assertEqual(ev_saved['LOCATION'], LOCATION)


    #@unittest.skip
    def test_paste_10_event_alarm(self) -> None:
        # Paste event with alarm
        SHORT_DESC = 'Test event with alarm'
        DT_ST = datetime(2030,7,21,23,30)
        alm = iAlarm()
        alm.add('TRIGGER', timedelta(minutes=-10))
        alm.add('ACTION', 'AUDIO')
        ev = self._new_event(SHORT_DESC, DT_ST, alarm=alm)
        self.assertEqual(len(ev.walk('VALARM')), 1)
        DT_TARG = date(2030,7,22)
        self._do_paste_event(ev, DT_TARG)

        # Test no entry exists where event was pasted from
        l = Calendar.occurrence_list(date(2030,1,1), date(2030,7,22))
        self.assertEqual(len(l), 0)

        # Test pasted entry
        l = Calendar.occurrence_list(date(2030,1,1), date(2031,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1].date(), DT_TARG)
        self.assertEqual(l[0][1].time(), DT_ST.time())
        pasted = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DTSTART'].dt.date(), DT_TARG)
        self.assertEqual(pasted['DTSTART'].dt.time(), DT_ST.time())
        pasted_alms = pasted.walk('VALARM')
        self.assertEqual(len(pasted_alms), 1)
        pasted_alm = pasted_alms[0]
        self.assertEqual(pasted_alm['TRIGGER'].dt, timedelta(minutes=-10))
        self.assertEqual(pasted_alm['ACTION'], 'AUDIO')

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt.date(), DT_TARG)
        self.assertEqual(ev_saved['DTSTART'].dt.time(), DT_ST.time())
        saved_alms = ev_saved.walk('VALARM')
        self.assertEqual(len(saved_alms), 1)
        saved_alm = saved_alms[0]
        self.assertEqual(saved_alm['TRIGGER'].dt, timedelta(minutes=-10))
        self.assertEqual(saved_alm['ACTION'], 'AUDIO')


    @unittest.skip
    def test_paste_11_event_alarm_abs(self) -> None:
        # Paste event with alarm given as absolute time (not offset)
        # SKIP for the moment - absolute time alarms generally not supported
        SHORT_DESC = 'Test event with absolute alarm'
        DT_ST = datetime(2030,7,21,23,30)
        alm = iAlarm()
        alm.add('TRIGGER', DT_ST - timedelta(minutes=15))
        alm.add('ACTION', 'AUDIO')
        ev = self._new_event(SHORT_DESC, DT_ST, alarm=alm)
        self.assertEqual(len(ev.walk('VALARM')), 1)
        DT_TARG = date(2030,7,20)
        self._do_paste_event(ev, DT_TARG)

        # Test no entry exists where event was pasted from
        l = Calendar.occurrence_list(date(2030,7,21), date(2031,1,1))
        self.assertEqual(len(l), 0)

        # Test pasted entry
        l = Calendar.occurrence_list(date(2030,1,1), date(2031,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1].date(), DT_TARG)
        self.assertEqual(l[0][1].time(), DT_ST.time())
        pasted = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(ev, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DTSTART'].dt.date(), DT_TARG)
        self.assertEqual(pasted['DTSTART'].dt.time(), DT_ST.time())
        pasted_alms = pasted.walk('VALARM')
        self.assertEqual(len(pasted_alms), 1)
        pasted_alm = pasted_alms[0]
        self.assertEqual(pasted_alm['TRIGGER'].dt, datetime(2030,7,20,23,15))
        self.assertEqual(pasted_alm['ACTION'], 'AUDIO')

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt.date(), DT_TARG)
        self.assertEqual(ev_saved['DTSTART'].dt.time(), DT_ST.time())
        saved_alms = ev_saved.walk('VALARM')
        self.assertEqual(len(saved_alms), 1)
        saved_alm = saved_alms[0]
        self.assertEqual(saved_alm['TRIGGER'].dt, datetime(2030,7,20,23,15))
        self.assertEqual(saved_alm['ACTION'], 'AUDIO')


    #@unittest.skip
    def test_paste_12_todo_simple(self) -> None:
        # Paste simple todo
        SHORT_DESC = 'Test todo'
        td = self._new_todo(SHORT_DESC)
        self._do_paste_todo(td)

        # Test imported entry
        l = Calendar.todo_list()
        self.assertEqual(len(l), 1)
        pasted = l[0] # Simple list (not occs) for todos
        self.check_stamps(td, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)

        # Test version saved to disk is also OK
        td_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, td_saved, same_obj=True)
        self.assertEqual(td_saved['SUMMARY'], SHORT_DESC)


    #@unittest.skip
    def test_paste_13_todo_simple_cat(self) -> None:
        # Paste todo from one cat (list) to another
        SHORT_DESC = 'Test todo in cat1 paste to cat2'
        td = self._new_todo(SHORT_DESC, cats=('cat1',))
        self.assertEqual(td['CATEGORIES'].cats[0], 'cat1')
        self._do_paste_todo(td, targ_cats=('cat2',))

        # Test imported entry
        l = Calendar.todo_list()
        self.assertEqual(len(l), 1)
        pasted = l[0] # Simple list (not occs) for todos
        self.check_stamps(td, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['CATEGORIES'].cats[0], 'cat2')

        # Test version saved to disk is also OK
        td_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, td_saved, same_obj=True)
        self.assertEqual(td_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(td_saved['CATEGORIES'].cats[0], 'cat2')


    #@unittest.skip
    def test_paste_14_todo_priority_status(self) -> None:
        # Paste todo with priority & status
        SHORT_DESC = 'Test todo with priority & status'
        PRIORITY = 3
        STATUS = 'COMPLETED'
        td = self._new_todo(SHORT_DESC, priority=PRIORITY, status=STATUS)
        self._do_paste_todo(td)

        # Test pasted entry
        l = Calendar.todo_list()
        self.assertEqual(len(l), 1)
        pasted = l[0] # Simple list (not occs) for todos
        self.check_stamps(td, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['PRIORITY'], PRIORITY)
        self.assertEqual(pasted['STATUS'], STATUS)

        # Test version saved to disk is also OK
        td_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, td_saved, same_obj=True)
        self.assertEqual(td_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(td_saved['PRIORITY'], PRIORITY)
        self.assertEqual(td_saved['STATUS'], STATUS)


    #@unittest.skip
    def test_paste_15_todo_due_date(self) -> None:
        # Paste todo with due date
        SHORT_DESC = 'Test todo with due date'
        DT_DUE = date(1999,12,31)
        td = self._new_todo(SHORT_DESC, due=DT_DUE)
        self._do_paste_todo(td)

        # Test pasted entry
        l = Calendar.todo_list()
        self.assertEqual(len(l), 1)
        pasted = l[0] # Simple list (not occs) for todos
        self.check_stamps(td, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DUE'].dt, DT_DUE)

        # Test version saved to disk is also OK
        td_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, td_saved, same_obj=True)
        self.assertEqual(td_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(td_saved['DUE'].dt, DT_DUE)


    #@unittest.skip
    def test_paste_16_todo_longdesc(self) -> None:
        # Paste todo with longdesc
        SHORT_DESC = 'Test todo with long description'
        LONG_DESC = u"""

This description has some newlines and unicode emoji.

😉 😂 😋
"""
        td = self._new_todo(SHORT_DESC, longdesc=LONG_DESC)
        self._do_paste_todo(td)

        # Test pasted entry
        l = Calendar.todo_list()
        self.assertEqual(len(l), 1)
        pasted = l[0] # Simple list (not occs) for todos
        self.check_stamps(td, pasted, new_copy=True)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DESCRIPTION'], LONG_DESC)

        # Test version saved to disk is also OK
        td_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, td_saved, same_obj=True)
        self.assertEqual(td_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(td_saved['DESCRIPTION'], LONG_DESC)


    #@unittest.skip
    def test_paste_17_todo_as_event(self) -> None:
        # Paste todo as an event
        SHORT_DESC = 'Test todo (destined to become an event)'
        CATS = ('cat1',)
        DT_DUE = date(2024,3,5)
        PRIORITY = 8
        STATUS = 'NEEDS-ACTION'
        LONG_DESC = 'Long description here'
        td = self._new_todo(SHORT_DESC, cats=CATS, due=DT_DUE, priority=PRIORITY, status=STATUS, longdesc=LONG_DESC)
        self.assertEqual(td['SUMMARY'], SHORT_DESC)
        self.assertEqual(td['CATEGORIES'].cats[0], CATS[0])
        self.assertEqual(td['DUE'].dt, DT_DUE)
        self.assertEqual(td['PRIORITY'], PRIORITY)
        self.assertEqual(td['STATUS'], STATUS)
        self.assertEqual(td['DESCRIPTION'], LONG_DESC)

        DT_TARG = date(2029,7,1)
        self._do_paste_event(td, DT_TARG)

        # Test pasted entry
        l = Calendar.occurrence_list(date(2029,1,1), date(2030,1,1))
        self.assertEqual(len(l), 1)
        self.assertEqual(l[0][1], DT_TARG)
        pasted = l[0][0] # occurrences, so entry/datetime pair
        self.check_stamps(td, pasted, new_copy=True, same_type=False)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertEqual(pasted['DTSTART'].dt, DT_TARG)
        self.assertEqual(pasted['CATEGORIES'].cats, td['CATEGORIES'].cats)
        self.assertNotIn('DUE', pasted)
        self.assertEqual(pasted['PRIORITY'], PRIORITY)
        self.assertNotIn('STATUS', pasted)
        self.assertEqual(pasted['DESCRIPTION'], LONG_DESC)

        # Test version saved to disk is also OK
        ev_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, ev_saved, same_obj=True)
        self.assertEqual(ev_saved['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev_saved['DTSTART'].dt, DT_TARG)
        self.assertEqual(ev_saved['CATEGORIES'].cats, td['CATEGORIES'].cats)
        self.assertNotIn('DUE', ev_saved)
        self.assertEqual(ev_saved['PRIORITY'], PRIORITY)
        self.assertNotIn('STATUS', ev_saved)
        self.assertEqual(ev_saved['DESCRIPTION'], LONG_DESC)


    #@unittest.skip
    def test_paste_18_event_as_todo(self) -> None:
        # Paste event as a todo
        SHORT_DESC = 'Test event (destined to become a todo)'
        DT_ST = datetime(2024,5,11,17,5)
        DT_END = datetime(2024,5,11,18,55)
        STATUS = 'CONFIRMED'
        LONG_DESC = 'Some long description'
        LOCATION = '221b Baker Street, London, England'
        ev = self._new_event(SHORT_DESC, DT_ST, end=DT_END, status=STATUS, longdesc=LONG_DESC, location=LOCATION)
        self.assertEqual(ev['SUMMARY'], SHORT_DESC)
        self.assertEqual(ev['DTSTART'].dt, DT_ST)
        self.assertEqual(ev['DTEND'].dt, DT_END)
        self.assertEqual(ev['STATUS'], STATUS)
        self.assertEqual(ev['DESCRIPTION'], LONG_DESC)
        self.assertEqual(ev['LOCATION'], LOCATION)

        self._do_paste_todo(ev)

        # Test pasted entry
        l = Calendar.todo_list()
        self.assertEqual(len(l), 1)
        pasted = l[0]
        self.check_stamps(ev, pasted, new_copy=True, same_type=False)
        self.assertEqual(pasted['SUMMARY'], SHORT_DESC)
        self.assertNotIn('DTSTART', pasted)
        self.assertNotIn('DTEND', pasted)
        self.assertNotIn('DUE', pasted)
        self.assertNotIn('STATUS', pasted)
        self.assertEqual(pasted['DESCRIPTION'], LONG_DESC)
        self.assertEqual(pasted['LOCATION'], LOCATION)

        # Test version saved to disk is also OK
        td_saved = self._get_saved_version(pasted)
        self.check_stamps(pasted, td_saved, same_obj=True)
        self.assertEqual(td_saved['SUMMARY'], SHORT_DESC)
        self.assertNotIn('DTSTART', td_saved)
        self.assertNotIn('DTEND', td_saved)
        self.assertNotIn('DUE', td_saved)
        self.assertNotIn('STATUS', td_saved)
        self.assertEqual(td_saved['DESCRIPTION'], LONG_DESC)
        self.assertEqual(td_saved['LOCATION'], LOCATION)


    def check_stamps(self, en1, en2, same_obj=False, new_copy=False, same_type=True) -> None:
        # Helper function to test uids/timestamps of two entries consistent.
        # Assumes en2 derived from en1, so dtstamp of en2 is later.
        # Args:
        #    same_obj: Really should be identical (e.g. save to storage)
        #    new_copy: Split from old (e.g. by copy-pasting) so different
        if same_type:
            self.assertEqual(type(en1), type(en2))
        else:
            self.assertNotEqual(type(en1), type(en2))
        if new_copy:
            self.assertNotEqual(en1['UID'], en2['UID'])
            self.assertLess(en1['CREATED'].dt, en2['CREATED'].dt)
        else:
            self.assertEqual(en1['UID'], en2['UID'])
            self.assertEqual(en1['CREATED'].dt, en2['CREATED'].dt)
        if same_obj:
            self.assertEqual(en1['DTSTAMP'].dt, en2['DTSTAMP'].dt)
        else:
            self.assertLess(en1['DTSTAMP'].dt, en2['DTSTAMP'].dt)


# Run all tests if this file is executed as main
if __name__ == '__main__':
    unittest.main()
